import itertools
import operator

from abc import ABC

# TODO: symbolic formatting: __add__ -> "+"

BINARY_LEVELS = {
        "__add__": 2,
        "__sub__": 2,

        "__mul__": 1,
        "__matmul__": 1,
        "__truediv__": 1,
        "__floordiv__": 1,
        "__mod__": 1,
        "__divmod__": 1,

        "__pow__": 0,

        "__lshift__": 3,
        "__rshift__": 3,

        "__and__": 4,
        "__xor__": 4,
        "__or__": 4,
        "__gt__": 5,
        "__lt__": 5,
        "__eq__": 5,
        "__ne__": 5,
        "__ge__": 5,
        "__le__": "5",

        "__getattr__": 0,
        "__getitem__": 0,
        }

BINARY_OPS = {
        "__add__": "+",
        "__sub__": "-",
        "__mul__": "*",
        "__matmul__": "@",
        "__truediv__": "/",
        "__floordiv__": "//",
        "__mod__": "%",
        "__divmod__": "IDK",
        "__pow__": "**",
        "__lshift__": "<<",
        "__rshift__": ">>",
        "__and__": "&",
        "__xor__": "^",
        "__or__": "|",
        "__gt__": ">",
        "__lt__": "<",
        "__eq__": "==",
        "__ne__": "!=",
        "__ge__": ">=",
        "__le__": "<=",
        "__getattr__": ".",
        "__getitem__": "[",
        }

UNARY_OPS = {
        "__invert__": "~",
        "__neg__": "-",
        "__pos__": "+",
        "__abs__": "ABS => "
        }

# TODO: can it just be put in binary ops? Special handling in Symbolic class?
ALL_OPS = {**BINARY_OPS, **UNARY_OPS}

BINARY_RIGHT_OPS = {}
for k, v in BINARY_OPS.items():
    BINARY_RIGHT_OPS[k.replace("__", "__r", 1)] = k

for k, v in BINARY_LEVELS.copy().items():
    BINARY_LEVELS[k.replace("__", "__r", 1)] = v

class Formatter:
    def __init__(self): pass

    def format(self, call, pad = 0):
        """Return a Symbolic or Call back as a nice tree, with boxes for nodes."""

        fmt_block = "█─"
        fmt_pipe = "├─"
        
        # TODO: why are some nodes still symbolic?
        if isinstance(call, Symbolic):
            return self.format(strip_symbolic(call))

        if isinstance(call, MetaArg):
            return "_"

        if isinstance(call, Call):
            call_str = fmt_block + ALL_OPS.get(call.func, repr(call.func))

            args_str = [self.format(arg) for arg in call.args]

            # format keyword args, making sure "└─<key> = █─" aligns box's children
            kwargs_str = []
            for k, v in call.kwargs.items():
                kwargs_str.append(
                        k + " = " + self.format(v, pad = len(k) + 3)
                        )

            all_args = [*args_str, *kwargs_str]
            padded = []
            for ii, entry in enumerate(all_args):
                chunk = self.fmt_pipe(
                        entry,
                        final = ii == len(all_args) - 1,
                        pad = pad
                        )
                padded.append(chunk)

            return "".join([call_str, *padded])

        return repr(call)


    @staticmethod
    def fmt_pipe(x, final = False, pad = 0):
        if not final:
            connector = "│ " if not final else "  "
            prefix = "├─"
        else:
            connector = "  "
            prefix = "└─"

        connector = "\n" + " "*pad + connector
        prefix = "\n" + " "*pad + prefix
        # NOTE: because visiting is depth first, this is essentially prepending
        # the text to the left.
        return prefix + connector.join(x.splitlines())


# Calls
# =============================================================================

class Call:
    """
    Representation of python operations.

    This class is responsible for representing the pieces of a python expression,
    as a function, along with its args and kwargs.

    For example, "some_object.a" would be represented at the function "__getattr__",
    with the args `some_object`, and `"a"`.

    Args:
        func: name of the function called. Class methods are represented using the names
              they have when defined on the class.
        *args: arguments the func call uses.
        **kwargs: keyword arguments the func call uses.


    """
    def __init__(self, func, *args, **kwargs):
        self.func = func
        self.args = args
        self.kwargs = kwargs

    def __repr__(self):
        # TODO: format binary, unary, call, associative
        if self.func in BINARY_OPS:
            op_repr = BINARY_OPS[self.func]
            fmt = "({args[0]} {func} {args[1]})"
        elif self.func in UNARY_OPS:
            op_repr = UNARY_OPS[self.func]
            fmt = "({func}{args[0]})"
        elif self.func == "getattr":
            op_repr = "."
            fmt = "({args[0]}{func}{args[1]})"
        else:
            op_repr, *arg_str = map(repr, self.args)
            kwarg_str = (str(k) + " = " + repr(v) for k,v in self.kwargs.items())

            combined_arg_str = ",".join(itertools.chain(arg_str, kwarg_str))
            fmt = "{}({})".format(op_repr, combined_arg_str)
            return fmt

        return fmt.format(
                    func = op_repr or self.func,
                    args = self.args,
                    kwargs = self.kwargs
                    )

    def __call__(self, x):
        args, kwargs = self.map_subcalls(self.evaluate_calls, args = (x,))
        inst, *rest = args

        #inst, *rest = (self.evaluate_calls(arg, x) for arg in self.args)
        #kwargs = {k: self.evaluate_calls(v, x) for k, v in self.kwargs.items()}
        
        # TODO: temporary workaround, for when only __get_attribute__ is defined
        if self.func == "__getattr__":
            return getattr(inst, *rest)
        elif self.func == "__getitem__":
            return operator.getitem(inst, *rest)
        elif self.func == "__call__":
            return getattr(inst, self.func)(*rest, **kwargs)

        # in normal case, get method to call, and then call it
        f_op = getattr(operator, self.func)
        return f_op(inst, *rest, **kwargs)

    @staticmethod
    def evaluate_calls(arg, x):
        if isinstance(arg, Call): return arg(x)

        return arg

    def copy(self):
        args, kwargs = self.map_subcalls(lambda child: child.copy())
        return self.__class__(self.func, *args, **kwargs)

    def map_subcalls(self, f, args = tuple(), kwargs = None):
        if kwargs is None: kwargs = {}

        new_args = tuple(f(arg, *args, **kwargs) if isinstance(arg, Call) else arg for arg in self.args)
        new_kwargs = {k: f(v, *args, **kwargs) if isinstance(v, Call) else v for k,v in self.kwargs.items()}

        return new_args, new_kwargs

    def map_replace(self, f):
        args, kwargs = self.map_subcalls(f)
        return self.__class__(self.func, *args, **kwargs)

    def iter_subcalls(self, f):
        yield from iter(arg for arg in self.args if instance(arg, Call))
        yield from iter(v for k,v in self.kwargs.items() if isinstance(v, Call))

    def op_vars(self, attr_calls = True):
        """Return set of all variable names used in Call
        
        Args:
            attr_calls: whether to include called attributes (e.g. 'a' from _.a())
        """
        varnames = set()

        op_var = self._get_op_var()
        if op_var is not None:
            varnames.add(op_var)

        if (not attr_calls
            and self.func == "__call__"
            and isinstance(self.args[0], Call)
            and self.args[0].func == "__getattr__"
            ):
            # skip obj, since it fetches an attribute this node is calling
            prev_obj, prev_attr = self.args[0].args
            all_args = itertools.chain([prev_obj], self.args[1:], self.kwargs.values())
        else:
            all_args = itertools.chain(self.args, self.kwargs.values())


        for arg in all_args:
            if isinstance(arg, Call):
                varnames.update(arg.op_vars(attr_calls = attr_calls))

        return varnames
    
    def _get_op_var(self):
        if self.func in ("__getattr__", "__getitem__") and isinstance(self.args[1], str):
            return self.args[1]

    def obj_name(self):
        obj = self.args[0]
        if isinstance(obj, Call):
            if obj.func == "__getattr__":
                return obj.args[0]
        elif hasattr(obj, '__name__'):
            return obj.__name__

        return None


class Lazy(Call):
    def __init__(self, func, arg = None):
        if arg is None:
            self.func = "<lazy>"
            self.args = [func]
        else:
            # will happen in generic node calls, e.g. self.__class__(self.func, ...)
            self.func = func
            self.args = [arg]

        self.kwargs = {}

    def __call__(self, x, *args, **kwargs):
        return self.args[0]



class UnaryOp(Call):
    def __repr__(self):
        fmt = "{func}{args[0]}"

        func = UNARY_OPS[self.func]
        return fmt.format(func = func, args = self.args, kwargs = self.kwargs)


class BinaryOp(Call):

    def __repr__(self):
        return self._repr(reverse = False)


    def _repr(self, reverse = False):
        func_name = self.get_func_name()

        level = BINARY_LEVELS[func_name]
        spaces = "" if level == 0 else " "

        args = self.args
        arg0 = "({args[0]})" if self.needs_paren(args[0]) else "{args[0]}"
        arg1 = "({args[1]})" if self.needs_paren(args[1]) else "{args[1]}"

        # handle binary ops that are not infix operators
        if self.func == "__getitem__":
            suffix = "]"
        else:
            suffix = ""

        # final, formatting
        fmt = arg0 + "{spaces}{func}{spaces}" + arg1 + suffix


        func = BINARY_OPS[func_name]
        if self.func == "__getattr__":
            # use bare string. eg _.a
            fmt_args = [repr(args[0]), args[1]]
        else:
            fmt_args = list(map(repr, args))
        
        if reverse:
            fmt_args = list(reversed(fmt_args))

        return fmt.format(func = func, args = fmt_args, spaces = spaces)

    def get_func_name(self):
        return BINARY_RIGHT_OPS.get(self.func, self.func)
        

    def needs_paren(self, x):
        if isinstance(x, BinaryOp):
            sub_lvl = BINARY_LEVELS[x.get_func_name()]
            level = BINARY_LEVELS[self.get_func_name()]
            if sub_lvl != 0 and sub_lvl > level:
                return True

        return False

class BinaryRightOp(BinaryOp):

    def __call__(self, x):
        inst, *rest = (self.evaluate_calls(arg, x) for arg in self.args)
        kwargs = {k: self.evaluate_calls(v, x) for k, v in self.kwargs.items()}
        
        # in normal case, get method to call, and then call it
        func_name = BINARY_RIGHT_OPS[self.func]
        f_op = getattr(operator, func_name)

        # TODO: in practice rest only has 1 item, but this is not enforced..
        return f_op(*rest, inst, **kwargs)

    def __repr__(self):
        return self._repr(reverse = True)


class DictCall(Call):
    """evaluates both keys and vals."""

    def __init__(self, f, *args, **kwargs):
        # TODO: validation, clean up class
        super().__init__(f, *args, **kwargs)
        self.args = (dict, dict(self.args[1]))

    def map_subcalls(self, f, args = tuple(), kwargs = None):
        if kwargs is None: kwargs = {}

        d = self.args[1]

        # TODO: don't use a giant comprehension
        new_d = {
            f(k, *args, **kwargs) if isinstance(k, Call) else k     # key part
            : f(v, *args, **kwargs) if isinstance(v, Call) else v   # val part
            for k,v in d.items()
            }

        return (self.args[0], new_d), {}

    def __call__(self, x):
        # TODO: note that it will not descend into dict to evaluate calls
        return self.args[1]


# Slice Calls -----------------------------------------------------------------
# note the metaclass SliceOp below, which registers SliceOpExt

class _SliceOpExt(Call):
    """Represent arguments for extended slice syntax.

    E.g. expressions like _[_:, 'a', 1:2, ] use a tuple of many slices.
    """

    def __init__(self, func, *args, **kwargs):
        self.func = "__siu_slice__"

        if kwargs:
            raise ValueError("a slice cannot accept keyword arguments")

        self.args = args
        self.kwargs = {}

    def __repr__(self):
        return ", ".join(map(self._repr_slice, self.args))

    def __call__(self, x):
        args, kwargs = self.map_subcalls(self.evaluate_calls, args = (x,))

        return args

    def map_subcalls(self, f, args=tuple(), kwargs=None):
        if kwargs is None: kwargs = {}

        # evaluate each argument, which can be a slice
        args = tuple(self._apply_slice_entry(f, obj, args, kwargs) for obj in self.args)

        return args, {}

    def op_vars(self, *args, **kwargs):
        def visit_op_vars(node, *args, **kwargs):
            if not isinstance(node, Call):
                return set()

            return node.op_vars(*args, **kwargs)

        results = set()
        for child in self.args:
            # TODO: refactor Call classes to use singledispatch, so we can call
            # op_vars directly on a slice object
            if isinstance(child, slice):
                for piece in (child.start, child.stop, child.step):
                    results.update(visit_op_vars(piece, *args, **kwargs))

            else:
                results.update(visit_op_vars(child, *args, **kwargs))

        return results

    @staticmethod
    def _apply_slice_entry(f, slice_, args, kwargs):
        if not isinstance(slice_, slice):
            # a slice argument can be simple literal. e.g. _['a']
            return f(slice_, *args, **kwargs) if isinstance(slice_, Call) else slice_

        slice_args = (slice_.start, slice_.stop, slice_.step)
        args = [
            f(entry, *args, **kwargs) if isinstance(entry, Call) else entry
            for entry in slice_args
            ]

        return slice(*args)

    @staticmethod
    def _unslice(slice_):
        if isinstance(slice_, slice):
            return (slice_.start, slice_.stop, slice_.step)

        return slice_

    @classmethod
    def _repr_slice(cls, slice_):
        if isinstance(slice_, slice):
            pieces = cls._unslice(slice_)

            # if the last piece is not specified, e.g. :1, then we want to represent
            # it like that, and not the equivalent :1:.
            if pieces[-1] is None:
                pieces = pieces[:-1]
        else:
            pieces = (slice_,)

        return ":".join(repr(x) if x is not None else "" for x in pieces)


class _SliceOpIndex(_SliceOpExt):
    """Special case of slicing, where getitem receives a single slice object."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __call__(self, *args, **kwargs):
        res = super().__call__(*args, **kwargs)
        return res[0]


class SliceOp(ABC):
    """Factory class for creating slice calls. 

    Note that it has SliceOpIndex, and SliceOpExt registered as subclasses, so
    that this class can be used rather than the specific implementations.
    """

    def __new__(cls, func, *args, **kwargs):
        # must be constructed in the same way as __getitem__ sees the indexer.
        # that is, a single argument (that can be a tuple)
        if len(args) != 1:
            raise ValueError("SliceOpIndex allows 1 argument, but received %s" % len(args))

        elif isinstance(args[0], tuple):
            # general case, where calling returns a tuple of indexers.
            # e.g. _['a', ], or _[1:, :, :, :]
            return _SliceOpExt(func, *args[0])

        else:
            # special case, where calling returns a single indexer, rather than
            # a tuple of indexers (e.g. _['a'], or _[1:])
            return _SliceOpIndex(func, args[0])


SliceOp.register(_SliceOpExt)


# Special kinds of call arguments ----
# These functions insure that when using siu expressions generated by _,
# that call.args[0] is always another call. This allows them to trivially
# respond to calling a siu expression, map_subcalls, etc..
#
# In the future, could make a parent class for Call, with a restricted
# set of behavior similar to theirs.
#
# TODO: validate that call.args[0] is a Call in tree visitors?

class MetaArg(Call):
    def __init__(self, func, *args, **kwargs):
        self.func = "_"
        self.args = tuple()
        self.kwargs = {}

    def __repr__(self):
        return self.func

    def __call__(self, x):
        return x

class FuncArg(Call):
    def __init__(self, func, *args, **kwargs):
        self.func = '__custom_func__'

        if func == '__custom_func__':
            func = args[0]

        self.args = tuple([func])
        self.kwargs = {}

    def __repr__(self):
        return repr(self.args[0])

    def __call__(self, x):
        return self.args[0]


# Trees and Visitors ==========================================================
from .error import ShortException

class FunctionLookupError(ShortException): pass

class FunctionLookupBound:
    def __init__(self, msg):
        self.msg = msg

    def __call__(self):
        raise NotImplementedError(self.msg)


class CallVisitor:
    """
    A node visitor base class that walks the call tree and calls a
    visitor function for every node found.  This function may return a value
    which is forwarded by the `visit` method.

    Note: essentially a copy of ast.NodeVisitor
    """

    def visit(self, node):
        """Visit a node."""
        method = 'visit_' + node.func
        visitor = getattr(self, method, self.generic_visit)
        return visitor(node)

    def generic_visit(self, node):
        """Called if no explicit visitor function exists for a node."""
        
        node.map_subcalls(self.visit)

    @classmethod
    def quick_visitor(cls, visit_dict, node):
        """Class method to quickly define and run a custom visitor."""

        qv = type('QuickVisitor', cls, visit_dict)
        qv.visit(node)

class CallListener:
    """Generic listener. Each exit is called on a node's copy."""
    def enter(self, node):
        method = 'enter_' + node.func
        f_enter = getattr(self, method, self.generic_enter)

        return f_enter(node)

    def exit(self, node):
        method = 'exit_' + node.func
        f_exit = getattr(self, method, self.generic_exit)
        return f_exit(node)

    def generic_enter(self, node):
        args, kwargs = node.map_subcalls(self.enter)

        return self.exit(node.__class__(node.func, *args, **kwargs))

    def generic_exit(self, node):
        return node

    def enter_if_call(self, x):
        if isinstance(x, Call):
            return self.enter(x)

        return x

def get_attr_chain(node, max_n):
    # TODO: need to make custom calls their own Call class, then will not have to
    #       do these kinds of checks, since a __call__ will always be on a Call obj
    if not isinstance(node, Call):
        return [], node

    out = []
    ttl_n = 0
    crnt_node = node
    while ttl_n < max_n:
        if crnt_node.func != "__getattr__": break

        obj, attr = crnt_node.args
        out.append(attr)

        ttl_n += 1
        crnt_node = obj

    return list(reversed(out)), crnt_node


from inspect import isclass, isfunction
from typing import get_type_hints
from .utils import is_dispatch_func_subtype

class CallTreeLocal(CallListener):
    def __init__(
            self,
            local,
            call_sub_attr = None,
            chain_sub_attr = False,
            dispatch_cls = None,
            result_cls = None,
            call_props = None
            ):
        """
        Arguments:
            local: a dictionary mapping func_name: func, used to replace call expressions.
            call_sub_attr: a set of attributes signaling any subattributes are property
                           methods. Eg. {'dt'} to signify in _.dt.year, year is a property call.
            chain_sub_attr: whether to included the attributes in the above argument, when looking up
                           up a replacement for the property call. E.g. does local have a 'dt.year' entry.
            dispatch_cls: if custom calls are dispatchers, dispatch on this class. If none, use their name
                          to try and get their corresponding local function.
            result_cls: if custom calls are dispatchers, require their result annotation to be a subclass
                          of this class.
            call_props: property methods to potentially convert to local calls.
        """
        self.local = local
        self.call_sub_attr = set(call_sub_attr or [])
        self.chain_sub_attr = chain_sub_attr
        self.dispatch_cls = dispatch_cls
        self.result_cls = result_cls
        self.call_props = set(call_props or [])

    def translate(self, expr):
        """Return the translation of an expression.

        This method is meant to be a high-level entrypoint.

        """

        # note that by contract, don't need to strip symbolic
        return self.enter(strip_symbolic(expr))

    def dispatch_local(self, name):
        func = self.local[name]

        if hasattr(func, "dispatch") and self.result_cls is not None:
            return func.dispatch(self.result_cls)

        return func

    def create_local_call(self, name, prev_obj, cls, func_args = None, func_kwargs = None):
        # need call attr name (arg[0].args[1]) 
        # need call arg and kwargs
        func_args = tuple() if func_args is None else func_args
        func_kwargs = {} if func_kwargs is None else func_kwargs

        try:
            local_func = self.dispatch_local(name)
        except KeyError as err:
            raise FunctionLookupError("Missing translation for function call: %s"% name)

        if isinstance(local_func, FunctionLookupBound):
            raise FunctionLookupError(local_func.msg)
            #local_func()

        #if isclass(local_func) and issubclass(local_func, Exception):
        #    raise local_func

        return cls(
                "__call__",
                local_func,
                prev_obj,
                *func_args,
                **func_kwargs
                )

    def enter(self, node):
        # if no enter metthod for operators, like __invert__, try to get from local
        # TODO: want to only do this if func is the name of an infix op's method
        method = 'enter_' + node.func
        if not hasattr(self, method) and node.func in self.local:
            args, kwargs = node.map_subcalls(self.enter)
            return self.create_local_call(node.func, args[0], Call, args[1:], kwargs)

        return super().enter(node)

    def enter___getattr__(self, node):
        obj, attr = node.args
        if obj.func == "__getattr__" and obj.args[1] in self.call_sub_attr:
            prev_obj, prev_attr = obj.args

            # TODO: should always call exit?
            if self.chain_sub_attr:
                # use chained attribute to look up local function instead
                # e.g. dt.round, rather than round
                attr = prev_attr + '.' + attr
            return self.create_local_call(attr, prev_obj, Call)
        elif attr in self.call_props:
            return self.create_local_call(attr, obj, Call)

        return self.generic_enter(node)

    def enter___custom_func__(self, node):
        func = node(None)
        
        # TODO: not robust at all, need class for singledispatch? unique attr flag?
        if (hasattr(func, 'registry')
            and hasattr(func, 'dispatch')
            and self.dispatch_cls is not None
            ):
            # allow custom functions that dispatch on dispatch_cls
            f_for_cls = func.dispatch(self.dispatch_cls)

            if isinstance(f_for_cls, FunctionLookupBound):
                # TODO: this is a bit funky, since FLB raises when called
                raise FunctionLookupError(f_for_cls.msg)

            if (self.result_cls is None
                or is_dispatch_func_subtype(f_for_cls, self.dispatch_cls, self.result_cls)
                ):
                # matches return annotation type (or not required)
                return node.__class__(f_for_cls)
            
            raise FunctionLookupError(
                    "External function {name} can dispatch on the class {dispatch_cls}, but "
                    "must also have result annotation of (sub)type {result_cls}"
                        .format(
                            name = func.__name__,
                            dispatch_cls = self.dispatch_cls,
                            result_cls = self.result_cls
                            )
                    )

        # doesn't raise an error so we can look in locals for now
        # TODO: remove behavior, once all SQL dispatch funcs moved from locals
        return self.generic_enter(node)

    def enter___call__(self, node):
        """
        Overview:
             variables      _.x.method(1)         row_number(_.x, 1)
             ---------      -------------        --------------------
            
                            █─'__call__'         █─'__call__'                           
             obj            ├─█─.                ├─<function row_number
                            │ ├─█─.              ├─█─.                                  
                            │ │ ├─_              │ ├─_                                  
                            │ │ └─'x'            │ └─'x'                                
                            │ └─'method'         │                                   
                            └─1                  └─1
        """
        obj, *rest = node.args
        args = tuple(self.enter_if_call(child) for child in rest)
        kwargs = {k: self.enter_if_call(child) for k, child in node.kwargs.items()}

        attr_chain, target = get_attr_chain(obj, max_n = 2)
        if attr_chain:
            # want _.x.method() -> method(_.x), need to transform
            if attr_chain[0] in self.call_sub_attr:
                # e.g. _.dt.round()
                call_name = ".".join(attr_chain) if self.chain_sub_attr else attr_chain[-1]
                entered_target = self.enter_if_call(target)
            else:
                call_name = attr_chain[-1]
                entered_target = self.enter_if_call(obj.args[0])

        elif isinstance(obj, FuncArg) and self.dispatch_cls is None:
            # want function(_.x) -> new_function(_.x), has form
            call_name = obj.obj_name()
            # the first argument is basically "self"
            entered_target, *args = args
        else:
            # default to generic enter
            return self.generic_enter(node)

        return self.create_local_call(
                call_name, entered_target, node.__class__,
                args, kwargs
                )


# Also need NodeTransformer


# Symbolic
# =============================================================================

class Symbolic(object):
    def __init__(self, source = None, ready_to_call = False):
        self.__source = MetaArg("_") if source is None else source
        self.__ready_to_call = ready_to_call


    # allowed methods ----

    def __getattr__(self, x):
        # temporary hack working around ipython pretty.py printing
        #if x == "__class__": return Symbolic

        return Symbolic(BinaryOp(
                "__getattr__",
                self.__source,
                strip_symbolic(x)
                ))
                

    def __call__(self, *args, **kwargs):
        if self.__ready_to_call:
            return self.__source(*args, **kwargs)

        return create_sym_call(self.__source, *args, **kwargs)

    def __getitem__(self, x):
        return Symbolic(BinaryOp(
                "__getitem__",
                self.__source,
                slice_to_call(x),
                ),
                ready_to_call = True)

    
    def __invert__(self):
        if isinstance(self.__source, Call) and self.__source.func == "__invert__":
            return self.__source.args[0]
        else: 
            return self.__op_invert()


    def __op_invert(self):
        return Symbolic(UnaryOp('__invert__', self.__source), ready_to_call = True)


    # banned methods ----

    __contains__ = None
    __iter__ = None

    def __bool__(self):
        raise TypeError("Symbolic objects can not be converted to True/False, or used "
                        "with these keywords: not, and, or.")


    # representation ----
        
    def __repr__(self):
        return Formatter().format(self.__source)


def create_sym_call(source, *args, **kwargs):
    return Symbolic(Call(
            "__call__",
            strip_symbolic(source),
            *map(strip_symbolic, args),
            **{k: strip_symbolic(v) for k,v in kwargs.items()}
            ),
            ready_to_call = True)


def slice_to_call(x):

    # TODO: uses similar code to SliceOp. make a walk_slice function?
    def f_strip(s):
        if isinstance(s, slice):
            return slice(*map(strip_symbolic, (s.start, s.stop, s.step)))

        return strip_symbolic(s)


    if isinstance(x, tuple):
        arg = tuple(map(f_strip, x))
    else:
        arg = f_strip(x)

    return SliceOp("__siu_slice__", arg)


def str_to_getitem_call(x):
    return Call("__getitem__", MetaArg("_"), x)

    
def strip_symbolic(x):
    if isinstance(x, Symbolic):
        return x.__dict__["_Symbolic__source"]

    return x


def explain(symbol):
    """Print representation that resembles code used to create symbol."""
    if isinstance(symbol, Symbolic):
        return str(strip_symbolic(symbol))

    return str(symbol)


# symbolic dispatch wrapper ---------------------------------------------------

from functools import singledispatch, update_wrapper
import inspect

def _dispatch_not_impl(func_name):
    def f(x, *args, **kwargs):
        raise TypeError("singledispatch function {func_name} not implemented for type {type}"
                            .format(func_name = func_name, type = type(x))
                            )

    return f

def symbolic_dispatch(f = None, cls = object):
    """Return a generic dispatch function with symbolic data implementations.

    The function dispatches (Call or Symbolic) -> FuncArg.

    Examples:
      ::

        @symbolic_dispatch(int)
        def add1(x): return x + 1
        
        @add1.register(str)
        def _add1_str: return int(x) + 1
        
        
        @symbolic_dispatch
        def my_func(x): raise NotImplementedError("no default method")

    """
    if f is None:
        return lambda f: symbolic_dispatch(f, cls)

    # TODO: don't use singledispatch if it has already been done
    dispatch_func = singledispatch(f)

    if cls is not object:
        dispatch_func.register(cls, f)
        dispatch_func.register(object, _dispatch_not_impl(dispatch_func.__name__))


    @dispatch_func.register(Symbolic)
    def _dispatch_symbol(__data, *args, **kwargs):
        return create_sym_call(FuncArg(dispatch_func), strip_symbolic(__data), *args, **kwargs)

    @dispatch_func.register(Call)
    def _dispatch_call(__data, *args, **kwargs):
        # TODO: want to just create call, for now use hack of creating a symbolic
        #       call and getting the source off of it...
        return strip_symbolic(create_sym_call(FuncArg(dispatch_func), __data, *args, **kwargs))

    return dispatch_func


    
# Do some gnarly method setting -----------------------------------------------

def create_binary_op(op_name, left_op = True):
    def _binary_op(self, x):
        if left_op:
            node = BinaryOp(op_name, strip_symbolic(self), strip_symbolic(x))
        else:
            node = BinaryRightOp(op_name, strip_symbolic(self), strip_symbolic(x))

        return Symbolic(node, ready_to_call = True)
    return _binary_op

def create_unary_op(op_name):
    def _unary_op(self):
        node = UnaryOp(op_name, strip_symbolic(self))

        return Symbolic(node, ready_to_call = True)

    return _unary_op

for k, v in BINARY_OPS.items():
    if k in {"__getattr__", "__getitem__"}: continue
    rop = k.replace("__", "__r", 1)
    setattr(Symbolic, k, create_binary_op(k))
    setattr(Symbolic, rop, create_binary_op(rop, left_op = False))

for k, v in UNARY_OPS.items():
    if k != "__invert__":
        setattr(Symbolic, k, create_unary_op(k))

Lam = Lazy

_ = Symbolic()




import pandas as pd
import pkg_resources

# mtcars ----------------------------------------------------------------------
_fname = pkg_resources.resource_filename("siuba.data", "mtcars.csv")

mtcars = pd.read_csv(_fname)
mtcars.__doc__ = """
mtcars data.

Source: Henderson and Velleman (1981), Building multiple regression models interactively. Biometrics, 37, 391–411.

--- Original DataFrame docs below ---
""" + mtcars.__doc__


# cars ------------------------------------------------------------------------
cars = mtcars[["cyl", "mpg", "hp"]]


# cars_sql --------------------------------------------------------------------
import siuba.sql.utils as _sql_utils
from siuba.sql import LazyTbl as _LazyTbl
cars_sql = _LazyTbl(
        _sql_utils.mock_sqlalchemy_engine("postgresql"),
        "cars",
        ["cyl", "mpg", "hp"]
        )




import importlib

def get_dialect_translator(name):
    mod = importlib.import_module('siuba.sql.dialects.{}'.format(name))
    return mod.translator

def get_dialect_funcs(name):
    #dialect = engine.dialect.name
    mod = importlib.import_module('siuba.sql.dialects.{}'.format(name))
    return mod.funcs

def get_sql_classes(name):
    mod = importlib.import_module('siuba.sql.dialects.{}'.format(name))
    win_name = name.title() + "Column"
    agg_name = name.title() + "ColumnAgg"
    return {
            'window': getattr(mod, win_name),
            'aggregate': getattr(mod, agg_name)
            }


def mock_sqlalchemy_engine(dialect):
    """
    Create a sqlalchemy.engine.Engine without it connecting to a database.

    Examples
    --------

    ::
        from siuba.sql import LazyTbl
        from siuba import _, mutate, show_query

        engine = mock_sqlalchemy_engine('postgresql')
        tbl = LazyTbl(engine, 'some_table', ['x'])

        query = mutate(tbl, y = _.x + _.x)
        show_query(query)

    """
    from sqlalchemy.engine import Engine
    from sqlalchemy.dialects import registry

    dialect_cls = registry.load('postgresql')   
    return Engine(None, dialect_cls(), '')  


# Temporary fix for pandas bug (https://github.com/pandas-dev/pandas/issues/35484)
from pandas.io import sql as _pd_sql

class _FixedSqlDatabase(_pd_sql.SQLDatabase):
    def execute(self, *args, **kwargs):
        return self.connectable.execute(*args, **kwargs)


# Backwards compatibility for sqlalchemy 1.3 ----------------------------------

import re
import sqlalchemy

RE_VERSION=r"(?P<major>\d+)\.(?P<minor>\d+).(?P<patch>\d+)"  
SQLA_VERSION=tuple(map(int, re.match(RE_VERSION, sqlalchemy.__version__).groups()))

def is_sqla_12():
    return SQLA_VERSION[:-1] == (1, 2)

def is_sqla_13():
    return SQLA_VERSION[:-1] == (1, 3)


def _sql_select(columns, *args, **kwargs):
    from sqlalchemy import sql
    if is_sqla_12() or is_sqla_13():
        # use old syntax, where columns are passed as a list
        return sql.select(columns, *args, **kwargs)

    return sql.select(*columns, *args, **kwargs)


def _sql_column_collection(data, columns):
    from sqlalchemy.sql.base import ColumnCollection, ImmutableColumnCollection

    if is_sqla_12() or is_sqla_13():
        return ImmutableColumnCollection(data, columns)

    return ColumnCollection(list(data.items())).as_immutable()


def _sql_add_columns(select, columns):
    if is_sqla_12() or is_sqla_13():
        for column in columns:
            select = select.column(column)
        return select

    return select.add_columns(*columns)


def _sql_with_only_columns(select, columns):
    if is_sqla_12() or is_sqla_13():
        return select.with_only_columns(columns)

    return select.with_only_columns(*columns)



from IPython import get_ipython
from IPython.core.history import HistoryAccessor
from IPython.core.completer import cursor_to_position, _FakeJediCompletion
from typing import Iterable, Any
from siuba.siu import Symbolic

import jedi
import functools


def _jedi_matches(self, cursor_column:int, cursor_line:int, text:str) -> Iterable[Any]:
    """

    Return a list of :any:`jedi.api.Completions` object from a ``text`` and
    cursor position.

    Parameters
    ----------
    cursor_column : int
        column position of the cursor in ``text``, 0-indexed.
    cursor_line : int
        line position of the cursor in ``text``, 0-indexed
    text : str
        text to complete

    Debugging
    ---------

    If ``IPCompleter.debug`` is ``True`` may return a :any:`_FakeJediCompletion`
    object containing a string with the Jedi debug information attached.
    """
    df = _find_df_in_history(self.shell)
    if df and isinstance(self.shell.user_ns.get("_"), Symbolic):
        namespaces = [{**self.namespace, "_": self.namespace[df]}]
    else:
        namespaces = [self.namespace]

    if self.global_namespace is not None:
        namespaces.append(self.global_namespace)

    completion_filter = lambda x:x
    offset = cursor_to_position(text, cursor_line, cursor_column)
    # filter output if we are completing for object members
    if offset:
        pre = text[offset-1]
        if pre == '.':
            if self.omit__names == 2:
                completion_filter = lambda c:not c.name.startswith('_')
            elif self.omit__names == 1:
                completion_filter = lambda c:not (c.name.startswith('__') and c.name.endswith('__'))
            elif self.omit__names == 0:
                completion_filter = lambda x:x
            else:
                raise ValueError("Don't understand self.omit__names == {}".format(self.omit__names))

    interpreter = jedi.Interpreter(text[:offset], namespaces)
    try_jedi = True

    try:
        # find the first token in the current tree -- if it is a ' or " then we are in a string
        completing_string = False
        try:
            first_child = next(c for c in interpreter._get_module().tree_node.children if hasattr(c, 'value'))
        except StopIteration:
            pass
        else:
            # note the value may be ', ", or it may also be ''' or """, or
            # in some cases, """what/you/typed..., but all of these are
            # strings.
            completing_string = len(first_child.value) > 0 and first_child.value[0] in {"'", '"'}

        # if we are in a string jedi is likely not the right candidate for
        # now. Skip it.
        try_jedi = not completing_string
    except Exception as e:
        # many of things can go wrong, we are using private API just don't crash.
        if self.debug:
            print("Error detecting if completing a non-finished string :", e, '|')

    if not try_jedi:
        return []
    try:
        return filter(completion_filter, interpreter.complete(column=cursor_column, line=cursor_line + 1))
    except Exception as e:
        if self.debug:
            return [_FakeJediCompletion('Oops Jedi has crashed, please report a bug with the following:\n"""\n%s\ns"""' % (e))]
        else:
            return []


def _match_df_name(dfs, commands):
    if not dfs:
        return []

    for command in reversed(commands):
        exact_match = [df for df in dfs if df == command]
        if exact_match:
            return exact_match[0]

        method_match = [df for df in dfs if df + "." in command]
        if method_match:
            return method_match[0]

        assign_match = [df for df in dfs if command.startswith(df + " = ")]
        if assign_match:
            return assign_match[0]

        import_match = [df for df in dfs if "import " + df in command]
        if import_match:
            return import_match[0]

        in_expression = [df for df in dfs if df + "," in command or df + ")" in command]
        if in_expression:
            return in_expression[0]

        any_match = [df for df in dfs if df in command]
        if any_match:
            return any_match[0]

    # return one of the dataframes
    return dfs[0]


def _find_df_in_history(shell):
    dfs = shell.run_line_magic("who_ls", "DataFrame")

    history = HistoryAccessor()
    commands = [command for _, _, command in history.get_tail(include_latest=True)]

    return _match_df_name(dfs, commands)


def siuba_jedi_override(shell):
    shell.Completer._jedi_matches = functools.partial(_jedi_matches, shell.Completer)


shell = get_ipython()

if shell is not None:
    siuba_jedi_override(shell)



# sqlvariant, allow defining 3 namespaces to override defaults
from ..translate import (
        SqlColumn, SqlColumnAgg, extend_base, win_agg,
        SqlTranslator
        )

from .base import base_scalar, base_agg, base_nowin
import sqlalchemy.sql.sqltypes as sa_types
from sqlalchemy import sql

# Custom dispatching in call trees ============================================

class SqliteColumn(SqlColumn): pass
class SqliteColumnAgg(SqlColumnAgg, SqliteColumn): pass

scalar = extend_base(
        base_scalar,
        )

aggregate = extend_base(
        base_agg
        )

window = extend_base(
        # TODO: should check sqlite version, since < 3.25 can't use windows
        base_nowin,
        sd = win_agg("stddev")
        )


translator = SqlTranslator.from_mappings(
        scalar, window, aggregate,
        SqliteColumn, SqliteColumnAgg
        )



import pandas as pd
import numpy as np

from typing import Union, Tuple, Dict, Optional
from pandas.core.groupby import DataFrameGroupBy

from siuba.dply.verbs import singledispatch2, gather, var_create, var_select


def pivot_longer_spec(data,
                      spec,
                      names_repair: Optional[str] = "check_unique",
                      values_drop_na: bool = False):
    raise NotImplementedError("TODO: see https://github.com/machow/siuba/issues/293")


@singledispatch2(pd.DataFrame)
def pivot_longer(
        __data,
        *args,
        names_to: Union[str, Tuple[str, ...]] = "name",
        names_prefix: Optional[str] = None,
        names_sep: Optional[str] = None,
        names_pattern: Optional[str] = None,
        names_ptypes: Optional[Tuple] = None,
        names_repair: str = "check_unique",
        values_to: str = "value",
        values_drop_na: bool = False,
        values_ptypes: Optional[Union[str, Tuple[str, ...]]] = None,
        values_transform: Optional[Dict] = dict(),
        ):
    
    if names_sep is not None and names_pattern is not None:
        raise ValueError("You may only use either `names_sep` or "
                         "`names_pattern`.")

    if isinstance(names_to, str):
        names_to = (names_to,)

    # Copied selection over from gather, maybe this can be compartmentalised?
    var_list = var_create(*args)
    od = var_select(__data.columns, *var_list)

    value_vars = list(od) or None

    id_vars = [col for col in __data.columns if col not in od]

    keep_data = __data.loc[:,id_vars]
    if value_vars is None:
        # While stack works in this case, it will later on merge in to the
        # original dataframe. To copy tidyr behaviour, we need to raise a
        # ValueError
        # stacked = __data.stack(dropna=values_drop_na)
        raise ValueError("Please provide at least 1 column or all columns "
                         "(shorthand: _[:]).")
    elif names_sep is not None or names_pattern is not None:
        to_stack = __data.loc[:,value_vars]
        column_index = (
            to_stack.columns.str.split(names_sep).map(tuple)
            if names_sep is not None
            # Split by names_pattern, and remove empty strings using filter
            else to_stack.columns.str.split(names_pattern).map(
                lambda x: tuple(list(filter(None, x)))
            )
        )
        split_lengths = np.array(column_index.map(len))

        if not np.all(split_lengths == split_lengths[0]):
            raise ValueError(
                    "Splitting by {} leads to unequal lenghts ({}).".format(
                        names_sep if names_sep is not None else names_pattern
                    )
                )
        
        if split_lengths[0] != len(names_to):
            raise ValueError("Splitting provided more values than provided in "
                             "`names_to`")
        
        # TODO: To set names for the new index, we need to feed in a list.
        # There's no particular reason to use a tuples as input in the first
        # place, might be worth reconsidering the choice of input format?
        # TODO: What if we don't use '_value' in the tuple? Need to check tidyr
        stack_idx = (
            [i for i, x in enumerate(list(names_to)) if x != "_value"]
            if names_to != ('_value',)
            else -1
        )
        names_to = [x if x != "_value" else None for x in names_to]

        column_index = column_index.set_names(names_to)

        to_stack.columns = column_index
        stacked = to_stack.stack(stack_idx)
        stacked = stacked.reset_index(level=stacked.index.nlevels - 1)

        if stack_idx == -1:
            stacked = stacked.drop(columns='level_1')
        if np.nan in names_to:
            stacked = stacked.drop(columns=[np.nan])
        if values_drop_na:
            stacked = stacked.dropna(axis = 1)
    else:
        stacked = __data.loc[:,value_vars].stack(dropna=values_drop_na)
        # Set column names for stack
        # As in tidyr `values_to` is ignored if `names_sep` or `names_pattern`
        # is provided.
        stacked.index.rename(names_to[0], level=1, inplace=True)
        stacked.name = values_to

    # values_transform was introduced in tidyr 1.1.0
    if values_to in values_transform:
        # TODO: error handling -- this won't work for dictionaries
        # list needs special handling, as it can only be applied to iterables,
        # not integers.
        if values_transform[values_to] == list:
            stacked = stacked.apply(lambda x: [x])
        else:
            stacked = stacked.apply(lambda x: values_transform[values_to](x))

    stacked_df = (
        # if `names_sep` or `names_pattern` are not provided `stacked` will
        # be a pd.Series and needs its index reset.
        stacked.reset_index(1)
        if names_sep is None and names_pattern is None
        else stacked
    )

    # If we want to pivot all but one, we are left with a `pd.Series`.
    # This needs to be converted to a DataFrame to serve as left element in a
    # merge
    if isinstance(keep_data, pd.Series):
        output_df = keep_data.to_frame().merge(stacked_df, left_index=True, right_index=True)
    elif keep_data.empty:
        output_df = stacked_df
    else:
        output_df = keep_data.merge(stacked_df, left_index=True, right_index=True)
    
    return output_df

@pivot_longer.register(DataFrameGroupBy)
def _pivot_longer_gdf(__data, *args, **kwargs):
    # TODO: consolidate all verbs that punt to DataFrame version (#118)
    prior_groups = [el.name for el in __data.grouper.groupings]

    df = __data.obj
    res = pivot_longer(df, *args, **kwargs)

    missing_groups = set(prior_groups) - set(res.columns)
    if missing_groups:
        raise ValueError(
                "When using pivot_longer on grouped data, the result must contain "
                "original grouping columns. Missing group columns: %s" %missing_groups
                )

    return res.groupby(prior_groups)



import json
import yaml
import pkg_resources
import pandas as pd

from siuba.ops import ALL_OPS

from siuba.siu import FunctionLookupBound
from siuba.sql.utils import get_dialect_translator

SQL_BACKENDS = ["postgresql", "redshift", "sqlite", "mysql", "bigquery"]
ALL_BACKENDS = SQL_BACKENDS + ["pandas"]

methods = pd.DataFrame(
        [{"full_name": k, **vars(f.operation)} for k, f in ALL_OPS.items()]
        )


# SQL spec construction -------------------------------------------------------

def read_dialect(name):
    translator = get_dialect_translator(name)

    support = []
    for k in ALL_OPS:
        support.append(read_sql_op(k, name, translator))

    df_support = pd.DataFrame(support)

    return df_support



def read_sql_op(name, backend, translator):
    f_win = translator.window.local.get(name)
    f_agg = translator.aggregate.local.get(name)

    # check FunctionLookupBound, a sentinal class for not implemented funcs
    support =  not (f_win is None or isinstance(f_win, FunctionLookupBound))
    metadata = getattr(f_win, "operation", {})

    # window functions should be a superset of agg functions
    if f_win is None and f_agg is not None:
        raise Exception("agg functions in %s without window funcs: %s" %(backend, name))

    if support and isinstance(f_agg, FunctionLookupBound):
        flags = "no_mutate"
    else:
        flags = ""

    meta = {"is_supported": support, "flags": flags, **metadata}

    return {"full_name": name, "backend": backend, "metadata": meta}


# pandas spec construction ----------------------------------------------------

def read_pandas_ops():
    from siuba.experimental.pd_groups.groupby import SeriesGroupBy

    all_meta = []
    for k, v in ALL_OPS.items():
        has_impl = not isinstance(v.dispatch(SeriesGroupBy), FunctionLookupBound)
        all_meta.append({
            "full_name": k,
            "metadata": dict(is_supported = has_impl),
            "backend": "pandas"}
            )

    return pd.DataFrame(all_meta)

# process examples ------------------------------------------------------------
# examples currently have form like _ + _, or _.corr(_)
# these need to be converted to _.x + _.y, and _x.corr(_.y)

def enrich_spec_entry(entry):
    from siuba.siu import _, strip_symbolic
    accessors = ['str', 'dt', 'cat', 'sparse']
    expr = strip_symbolic(eval(entry["example"], {"_": _}))

    accessor = [ameth for ameth in accessors if ameth in expr.op_vars()] + [None]

    tmp = {
            **entry,
            'is_property': expr.func == "__getattr__",
            'expr_frame': replace_meta_args(expr, _.x, _.y, _.z),
            'expr_series': expr,
            'accessor': accessor[0],
            }

    return tmp

def replace_meta_args(call, *args):
    from siuba.siu import strip_symbolic, MetaArg
    replacements = map(strip_symbolic, args)
    
    def replace_next(node):
        # swap in first replacement for meta arg (when possible)
        if isinstance(node, MetaArg):
            return next(replacements)
        
        # otherwise, recurse into node
        args, kwargs = node.map_subcalls(replace_next)

        # return a copy of node
        return node.__class__(node.func, *args, **kwargs)

    return replace_next(call)


# Main ========================================================================

sql_methods    = pd.concat(list(map(read_dialect, SQL_BACKENDS)))
pandas_methods = pd.DataFrame(read_pandas_ops())

wide_backends = (
        pd.concat([sql_methods, pandas_methods])
        .pivot("full_name", "backend", "metadata")
        )

full_methods = methods.merge(wide_backends, how = "left", on = "full_name")

# replace NA entries--but pandas' fillna has custom behavior around dicts, so
# need to use this terrible hack
def set_default_support(d):
    if d.get("is_supported"):
        return {**d, "support": "supported"}

    return {**d, "is_supported": False, "support": "maydo"}

for be_name in ALL_BACKENDS:
    full_methods[be_name] = full_methods[be_name].apply(set_default_support)


# Nest backends from each being a column, to a single column of dicts
df_spec = (full_methods
        .assign(backends = lambda _: _[ALL_BACKENDS].to_dict(orient = "records"))
        .drop(columns = ALL_BACKENDS)
        )

fname_spec = pkg_resources.resource_filename("siuba.ops.support", "examples.yml")
with open(fname_spec, "r") as f:
    orig_spec = yaml.safe_load(f)

df_spec["example"] = df_spec.full_name.apply(lambda x: orig_spec[x])

# roundtrip through JSON to get rid of custom numpy types
# probably a much better way to do this
json_spec = df_spec.set_index("full_name").to_json(orient = "index")
raw_spec = json.loads(json_spec)

spec = {k: enrich_spec_entry(entry) for k, entry in raw_spec.items()} 





"""
This module holds default translations from pandas syntax to sql for 3 kinds of operations...

1. scalar - elementwise operations (e.g. array1 + array2)
2. aggregation - operations that result in a single number (e.g. array1.mean())
3. window - operations that do calculations across a window
            (e.g. array1.lag() or array1.expanding().mean())

It is broken into 5 sections:

* Warnings
* Column classes
* Custom over clauses
* Translation function generators (e.g. sql_agg("stdev"))
* SqlTranslator for translating symbolic expressions.

"""

from sqlalchemy import sql

from siuba.siu import FunctionLookupBound

# warning for when sql defaults differ from pandas ============================
import warnings


class SiubaSqlRuntimeWarning(UserWarning): pass

def warn_arg_default(func_name, arg_name, arg, correct):
    warnings.warn(
            "\n{func_name} sql translation defaults "
            "{arg_name} to {arg}. To return identical result as pandas, use "
            "{arg_name} = {correct}.\n\n"
            "This warning only displays once per function".format(
                func_name = func_name, arg_name = arg_name, arg = repr(arg), correct = repr(correct)
                ),
            SiubaSqlRuntimeWarning
            )


# Column data types ===========================================================

from sqlalchemy.sql.elements import ColumnClause

class SqlBase(ColumnClause): pass

class SqlColumn(SqlBase): pass

class SqlColumnAgg(SqlBase): pass


# Custom over clause handling  ================================================

from sqlalchemy.sql.elements import Over


class CustomOverClause(Over):
    """Base class for custom window clauses in SQL translation."""

    def set_over(self, group_by, order_by):
        raise NotImplementedError()

    @classmethod
    def func(cls, name):
        raise NotImplementedError()


class AggOver(CustomOverClause):
    """Over clause for uses of functions min, max, avg, that return one value.

    Note that this class does not set order by, which is how these functions
    generally become their cumulative versions.
    """

    def set_over(self, group_by, order_by = None):
        self.partition_by = group_by
        return self

    @classmethod
    def func(cls, name):
        sa_func = getattr(sql.func, name)
        def f(col, *args, **kwargs) -> AggOver:
            return cls(sa_func(col, *args, **kwargs))

        return f


class RankOver(CustomOverClause): 
    """Over clause for ranking functions.

    Note that in python we might call rank(col), but in SQL the ranking column
    is defined using order by.
    """
    def set_over(self, group_by, order_by = None):
        crnt_partition = getattr(self.partition_by, 'clauses', tuple())
        self.partition_by = sql.elements.ClauseList(*crnt_partition, *group_by.clauses)
        return self

    @classmethod
    def func(cls, name):
        sa_func = getattr(sql.func, name)
        def f(col) -> RankOver:
            return cls(sa_func(), order_by = col)

        return f


class CumlOver(CustomOverClause):
    """Over clause for cumulative versions of functions like sum, min, max.

    Note that this class is also currently used for aggregates that might require
    ordering, like nth, first, etc..

    """
    def set_over(self, group_by, order_by):
        self.partition_by = group_by


        # do not override order by if it was set by the user. this might happen
        # in functions like nth, which gives the option to set it.
        if self.order_by is None or not len(self.order_by):
            if not len(order_by):
                warnings.warn(
                        "No order by columns explicitly set in window function. SQL engine"
                        "does not guarantee a row ordering. Recommend using an arrange beforehand.",
                        RuntimeWarning
                        )

            self.order_by = order_by


        return self

    @classmethod
    def func(cls, name, rows=(None, 0)):
        sa_func = getattr(sql.func, name)
        def f(col, *args, **kwargs) -> CumlOver:
            return cls(sa_func(col, *args, **kwargs), rows = rows)

        return f

# convenience aliases for class methods above
win_agg = AggOver.func
win_over = RankOver.func
win_cumul = CumlOver.func


# Simple clause translation generators ========================================

def sql_agg(name):
    sa_func = getattr(sql.func, name)
    return lambda col: sa_func(col)

def sql_scalar(name):
    sa_func = getattr(sql.func, name)
    return lambda col, *args: sa_func(col, *args)

def sql_colmeth(meth, *outerargs):
    def f(col, *args) -> SqlColumn:
        return getattr(col, meth)(*outerargs, *args)
    return f

def set_agg(name):
    # Ordered and theoretical set aggregates
    sa_func = getattr(sql.func, name)
    return lambda col, *args: sa_func(*args).within_group(col)

# Handling not implemented translations ----

def sql_not_impl(msg = ""):

    return FunctionLookupBound(msg or "function not implemented")

def win_absent(name):
    # Return an error, that is picked up by the translator.
    # this allows us to report errors at translation, rather than call time.
    return FunctionLookupBound("SQL dialect does not support window function {}.".format(name))

# Annotations ----
# these functions wrap translation generators, in order to provide metadata
# for running, e.g., unit tests. This is important in cases where the SQL 
# output does not match exactly what pandas does (for example, it returns a 
# float, when pandas returns an int).

def annotate(f = None, **kwargs):
    # allow it to work as a decorator
    if f is None:
        return lambda f: annotate(f, **kwargs)

    if hasattr(f, "operation"):
        raise ValueError("function already has an operation attribute")

    f.operation = kwargs

    return f

def wrap_annotate(f, **kwargs):
    from functools import wraps

    @wraps(f)
    def wrapper(*args, **kwargs):
        return f(*args, **kwargs)

    wrapper.operation = kwargs

    return wrapper


#  Translator =================================================================

def extend_base(mapping, **kwargs):
    return {**mapping, **kwargs}

from siuba.ops.translate import create_pandas_translator

# TODO: should inherit from a ITranslate class (w/ abstract translate method)
class SqlTranslator:
    """Translates symbolic column operations to sqlalchemy clauses.

    Note that SqlTranslator's main job is to hold two actual translators--
    one for windowed and one for aggregation contexts--and to call the
    translate method to the correct one.


    Example::

       from siuba.sql.translate import (
           SqlTranslator,
           SqlColumn, SqlColumnAgg,
           AggOver, sql_not_impl
           )

       base   = {"__add__": sql_colmeth("add")}
       window = {"mean":    AggOver.func("mean")}
       agg    = {"mean":    sql_not_impl("group by mean")}

       trans = SqlTranslator.from_mappings(
               base, window, agg,
               SqlColumn, SqlColumnAgg
               )

       from siuba import _
       trans.translate(_ + _.mean())
       trans.translate(_ + _.mean(), window = False)
   """

    def __init__(self, window, aggregate):
        self.window = window
        self.aggregate = aggregate

    def translate(self, expr, window = True):
        if window:
            return self.window.translate(expr)

        return self.aggregate.translate(expr)

    def from_mappings(base, window, aggregate, WinCls, AggCls):
        trans_win = {**base, **window}
        trans_agg = {**base, **aggregate}

        return SqlTranslator(
                window = create_pandas_translator(trans_win, WinCls, sql.elements.ClauseElement),
                aggregate = create_pandas_translator(trans_agg, AggCls, sql.elements.ClauseElement)
                )




# sqlvariant, allow defining 3 namespaces to override defaults
from ..translate import (
        SqlColumn, SqlColumnAgg, extend_base, win_agg,
        SqlTranslator, sql_not_impl
        )

from .base import base_scalar, base_agg, base_win

import sqlalchemy.sql.sqltypes as sa_types

from sqlalchemy import sql
from sqlalchemy.sql import func as fn

# Custom dispatching in call trees ============================================

class MysqlColumn(SqlColumn): pass
class MysqlColumnAgg(SqlColumnAgg, MysqlColumn): pass

def sql_str_strip(left = True, right = True):
    def f(col):
        # see https://stackoverflow.com/a/6858168/1144523
        lstrip = "^[[:space:]]+" if left else ""
        rstrip = "[[:space:]]+$" if right else ""
    
        or_op = "|" if lstrip and rstrip else ""
        regex = "(" + lstrip + or_op + rstrip + ")"
    
        return fn.regexp_replace(col, regex, "")

    return f

def sql_func_extract_dow_monday(col):
    # MYSQL: sunday starts, equals 1 (an int)
    # pandas: monday starts, equals 0 (also an int)

    raw_dow = fn.dayofweek(col)

    # monday is 2 in MYSQL, so use monday + 5 % 7
    return (raw_dow + 5) % 7

def sql_is_date_offset(period, is_start = True):

    # will check against one day in the past for is_start, v.v. otherwise
    fn_add = fn.date_sub if is_start else fn.date_add

    def f(col):
        get_period = getattr(fn, period)
        src_per = get_period(col)
        incr_per = get_period(fn_add(col, sql.text("INTERVAL 1 DAY")))

        return src_per != incr_per

    return f

def sql_func_truediv(x, y):
    return sql.cast(x, sa_types.Numeric()) / y

def sql_func_floordiv(x, y):
    return x.op("DIV")(y)

def sql_func_between(col, left, right, inclusive=True):
    if not inclusive:
        raise NotImplementedError("between must be inclusive")

    # TODO: should figure out how sqlalchemy prefers to set types, rather
    # than setting manually on this expression
    expr = col.between(left, right)
    expr.type = sa_types.Boolean()
    return expr

scalar = extend_base(
        base_scalar,

        # copied from postgres. MYSQL does true division over ints by default,
        # but it does not produce double precision.
        __div__ = sql_func_truediv,
        div = sql_func_truediv,
        divide = sql_func_truediv,
        rdiv = lambda x,y: sql_func_truediv(y, x),
        __rdiv__ = lambda x, y: sql_func_truediv(y, x),

        __truediv__ = sql_func_truediv,
        truediv = sql_func_truediv,
        __rtruediv__ = lambda x, y: sql_func_truediv(y, x),

        __floordiv__ = sql_func_floordiv,
        __rfloordiv__ = lambda x, y: sql_func_floordiv(y, x),

        between = sql_func_between,

        **{
          "str.lstrip": sql_str_strip(right = False),
          "str.rstrip": sql_str_strip(left = False),
          "str.strip": sql_str_strip(),
          "str.title": sql_not_impl()                       # see https://stackoverflow.com/q/12364086/1144523
        },
        **{
          "dt.dayofweek": sql_func_extract_dow_monday,
          "dt.dayofyear": lambda col: fn.dayofyear(col),
          "dt.days_in_month": lambda col: fn.dayofmonth(fn.last_day(col)),
          "dt.daysinmonth": lambda col: fn.dayofmonth(fn.last_day(col)),
          "dt.is_month_end": lambda col: col == fn.last_day(col),
          "dt.is_month_start": lambda col: fn.dayofmonth(col) == 1,
          "dt.is_quarter_start": sql_is_date_offset("QUARTER"),
          "dt.is_year_start": sql_is_date_offset("YEAR"),
          "dt.is_year_end": sql_is_date_offset("YEAR", is_start = False),
          # see https://dev.mysql.com/doc/refman/8.0/en/date-and-time-functions.html#function_week
          "dt.week": lambda col: fn.week(col, 1),
          "dt.weekday": sql_func_extract_dow_monday,
          "dt.weekofyear": lambda col: fn.week(col, 1),
        }
        )

aggregate = extend_base(
        base_agg
        )

window = extend_base(
        base_win,
        sd = win_agg("stddev")
        )

funcs = dict(scalar = scalar, aggregate = aggregate, window = window)

translator = SqlTranslator.from_mappings(
        scalar, window, aggregate,
        MysqlColumn, MysqlColumnAgg
        )




"""
Note: this test file was heavily influenced by its dbplyr counterpart.

https://github.com/tidyverse/tidyr/blob/master/tests/testthat/test-pivot-long.R
"""

# TODO: pivot_longer rows in different order from gather. Currently using
#       a row order invariant form of test. Should fix to conform to pivot_longer.
# TODO: need to be careful about index. Should preserve original indices for rows.
# TODO: Current behaviour of pivot_longer is to preserve the index. Therefore
#       it needs to be reset for any assertion.

from . import pivot_longer, pivot_longer_spec

import pytest
import pandas as pd
import numpy as np

from siuba.siu import Symbolic
from siuba import group_by

from siuba.tests.helpers import data_frame, assert_frame_sort_equal
from pandas.testing import assert_frame_equal, assert_series_equal


_ = Symbolic()


def test_pivot_all_cols_to_long():
    "can pivot all cols to long"

    src = data_frame(x = [1,2], y = [3,4])
    dst = data_frame(name = ["x", "y", "x", "y"], value = [1, 3, 2, 4])
    
    res = pivot_longer(src, _["x":"y"])

    assert_frame_equal(res.reset_index(drop=True), dst)


def test_values_interleaved_correctly():
    # TODO: fix order issue
    df = data_frame(x = [1,2], y = [10, 20], z = [100, 200])

    pv = pivot_longer(df, _[0:3])
    assert pv["value"].tolist() == [1, 10, 100, 2, 20, 200]


@pytest.mark.xfail
def test_spec_add_multi_columns():
    df = data_frame(x = [1,2], y = [3,4])

    # TODO: is this the right format for a spec
    sp = data_frame(_name = ["x", "y"], _value = "v", a = 1, b = 2)
    pv = pivot_longer_spec(df, spec = sp)

    assert pv.columns.tolist() == ["a", "b", "v"]


def test_preserves_original_keys():
    df = data_frame(x = [1,2], y = [2,2], z = [1,2])
    pv = pivot_longer(df, _["y":"z"])

    assert pv.columns.tolist() == ["x", "name", "value"]
    assert_series_equal(
        pv["x"],
        pd.Series(df["x"].repeat(2))
        )


def test_can_drop_missing_values():
    df = data_frame(x = [1, np.nan], y = [np.nan, 2])
    pv = pivot_longer(df, _["x":"y"], values_drop_na=True)

    assert pv["name"].tolist() == ["x", "y"]
    assert pv["value"].tolist() == [1, 2]


def test_can_handle_missing_combinations():
    df = data_frame(id = ["A", "B"], x_1 = [1, 3], x_2 = [2, 4], y_2 = ["a", "b"])
    pv = pivot_longer(df, -_.id, names_to = ("_value", "n"), names_sep = "_")

    pv_expected = pd.Series([np.nan, "a", np.nan, "b"],
                            index = [0, 0, 1, 1],
                            name = 'y')

    assert pv.columns.tolist() == ["id", "n", "x", "y"]
    assert pv["x"].tolist() == [1, 2, 3, 4]
    pd.testing.assert_series_equal(pv["y"], pv_expected)


@pytest.mark.xfail
def test_mixed_columns_are_auto_coerced():
    # TODO: pandas stack (and melt) coerces categorical data when stacking.
    df = data_frame(x = pd.Categorical(["a"]), y = pd.Categorical(["b"]))
    pv = pivot_longer(df, _["x":"y"])

    assert_series_equal(pv["value"], pd.Categorical(['a', 'b']))


def test_can_override_default_output_col_type():
    df = data_frame(x = "x", y = 1)
    pv = pivot_longer(df, _["x":"y"], values_transform = {"value": list})

    assert pv["value"].tolist() == [["x"], [1]]


@pytest.mark.xfail
def test_spec_can_pivot_to_multi_measure_cols():
    df = data_frame(x = "x", y = 1)
    sp = data_frame(_name = ["x", "y"], _value = ["X", "Y"], row = [1, 1])

    pv = pivot_longer_spec(df, sp)

    assert pv.columns.tolist() == ["row", "X", "Y"]
    assert pv["X"] == "x"
    assert pv["Y"] == 1


@pytest.mark.xfail
def test_original_col_order_is_preserved():
    df = data_frame(id = ["A", "B"],
        z_1 = [1, 7], y_1 = [2, 8], x_1 = [3, 9],
        z_2 = [4, 10], y_2 = [5, 11], x_2 = [6, 12]
    )
    pv = pivot_longer(df, -_.id, names_to = ("_value", "n"), names_sep = "_")

    assert pv.columns.tolist() == ["id", "n", "z", "y", "x"]


def test_handles_duplicate_column_names():
    # Cannot initiate data_frame with duplicate keys
    # df = data_frame(x = 1, a = 1, a = 2, b = 3, b = 4)
    df = pd.DataFrame.from_records(
        [(1, 1, 2, 3, 4)], columns = ["x", "a", "a", "b", "b"]
    )
    pv = pivot_longer(df, -_.x)

    assert pv.columns.tolist() == ["x", "name", "value"]
    assert pv["name"].tolist() == ["a", "a", "b", "b"]
    assert pv["value"].tolist() == [1, 2, 3, 4]


def test_can_pivot_duplicate_names_to_value():
    df = data_frame(x = 1, a_1 = 1, a_2 = 2, b_1 = 3, b_2 = 4)
    pv1 = pivot_longer(df, -_.x, names_to = ("_value", np.nan), names_sep = "_")
    pv2 = pivot_longer(df, -_.x, names_to = ("_value", np.nan), names_pattern = "(.)_(.)")
    pv3 = pivot_longer(df, -_.x, names_to = "_value", names_pattern = "(.)_.")

    assert pv1.columns.tolist() == ["x", "a", "b"]
    assert pv1["a"].tolist() == [1, 2]
    assert_frame_equal(pv2, pv1)
    assert_frame_equal(pv3, pv1)


def test_value_can_be_any_pos_in_names_to():
    samp = data_frame(
        i = np.arange(1, 5),
        y_t1 = np.random.standard_normal(4),
        y_t2 = np.random.standard_normal(4),
        z_t1 = [3] * 4,
        z_t2 = [-2] * 4,
    )

    value_first = pivot_longer(samp, -_.i,
                               names_to = ("_value", "time"), names_sep = "_")

    samp2 = samp.rename(columns={"y_t1": "t1_y", "y_t2": "t2_y",
                                 "z_t1": "t1_z", "z_t2": "t2_z"})
    
    value_second = pivot_longer(samp2, -_.i,
                                names_to = ("time", "_value"), names_sep = "_")
    
    assert_frame_equal(value_first, value_second)


def test_type_error_message_uses_var_names():
    # Error handling is by default 'better' in python than R
    # This test is tricky, as python doesn't care when stacking data of different types.
    df = data_frame(abc = 1, xyz = "b")
    try:
        # This should by default pivot everything, as with tidyr
        pivot_longer(df, _[:])
    except:
        # Ideally we'd print an error message here and compare if the keys are
        # printed correctly, but `pivot_longer` doesn't brake in python as in R
        # when stacking different data types. Not sure if we should 'make' it brake?
        print(err)


def test_grouping_is_preserved():
    # In siuba this actually tests 3 things:
    # 1) Can we pipe a grouped DataFrame in to `pivot_longer`? - Not yet
    # 2) Does it retain grouping? - Not yet
    # 3) Can we get the names of grouping _columns_ ie variables? - Not yet?
    df = data_frame(g = [1, 2], x1 = [1, 2], x2 = [3, 4])
    # Breaks; `pivot_longer` needs singledispatch for grouped DataFrames
    out = (
        df
        >> group_by(_.g)
        >> pivot_longer(_["x1":"x2"], names_to = "x", values_to = "v")
    )

    # Breaks, as group_vars does not exist yet.
    # For now, in pandas it is probably better to check if the DataFrame remains
    # grouped, and if it matches the expected output
    # assert group_vars(out) == "g"
    expected = data_frame(
        g = [1, 1, 2, 2],
        x = ["x1", "x2", "x1", "x2"],
        v = [1, 3, 2, 4],
        _index = [0, 0, 1, 1]
    ).groupby("g")

    # assert_frame_equal does not work with DataFrameGroupBy.
    isinstance(out, expected.__class__)
    assert_frame_equal(out.obj, expected.obj)



"""
Translations for the postgresql dialect of SQL.
"""

from ..translate import (
        win_agg, win_over, win_cumul, sql_scalar, sql_agg,
        RankOver,
        wrap_annotate, annotate,
        extend_base,
        SqlTranslator,
        )

from .base import (
        SqlColumn, SqlColumnAgg,
        base_scalar, base_win, base_agg
        )

import sqlalchemy.sql.sqltypes as sa_types
from sqlalchemy import sql


# Custom dispatching in call trees ============================================

class PostgresqlColumn(SqlColumn): pass

class PostgresqlColumnAgg(SqlColumnAgg, PostgresqlColumn): pass

# Custom translations =========================================================

def returns_float(ns, func_names):
    return {k: wrap_annotate(ns[k], result_type = "float") for k in func_names}

def sql_log(col, base = None):
    if base is None:
        return sql.func.ln(col)
    return sql.func.log(col)

@annotate(result_type = "float")
def sql_round(col, n):
    return sql.func.round(col, n)

def sql_func_contains(col, pat, case = True, flags = 0, na = None, regex = True):
    # TODO: warn there differences in regex for python and sql?
    # TODO: validate pat is string?
    if not isinstance(pat, str):
        raise TypeError("pat argument must be a string")
    if flags != 0 or na is not None:
        raise NotImplementedError("flags and na options not supported")

    if not regex:
        case_col = col if case else col.lower()
        return case_col.contains(pat, autoescape = True)

    full_op = "~" if case else "~*"

    return col.op(full_op)(pat)

def sql_func_truediv(x, y):
    return sql.cast(x, sa_types.Float()) / y


scalar = extend_base(
        base_scalar,

        # TODO: remove log, not a pandas method
        log = sql_log,

        # TODO: bring up to date (not pandas methods)
        concat = lambda col: sql.func.concat(col),
        cat = lambda col: sql.func.concat(col),
        str_c = lambda col: sql.func.concat(col),

        # infix and infix methods ----

        __div__ = sql_func_truediv,
        div = sql_func_truediv,
        divide = sql_func_truediv,
        rdiv = lambda x,y: sql_func_truediv(y, x),
        __rdiv__ = lambda x, y: sql_func_truediv(y, x),

        __truediv__ = sql_func_truediv,
        truediv = sql_func_truediv,
        __rtruediv__ = lambda x, y: sql_func_truediv(y, x),


        round = sql_round,
        __round__ = sql_round,

        **{
            "str.contains": sql_func_contains,
        },
        **returns_float(base_scalar, [
             "dt.day", "dt.dayofweek", "dt.dayofyear", "dt.days_in_month",
             "dt.daysinmonth", "dt.hour", "dt.minute", "dt.month",
             "dt.quarter", "dt.second", "dt.week", "dt.weekday",
             "dt.weekofyear", "dt.year"
             ]),
        )

window = extend_base(
        base_win,
        any = annotate(win_agg("bool_or"), input_type = "bool"),
        all = annotate(win_agg("bool_and"), input_type = "bool"),
        lag = win_agg("lag"),
        std = win_agg("stddev_samp"),
        var = win_agg("var_samp"),

        # overrides ----

        # note that postgres does sum(bigint) -> numeric
        size = win_agg("count"),     #TODO double check
        )

aggregate = extend_base(
        base_agg,
        all = sql_agg("bool_and"),
        any = sql_agg("bool_or"),
        std = sql_agg("stddev_samp"),
        var = sql_agg("var_samp"),
        )


funcs = dict(scalar = scalar, aggregate = aggregate, window = window)

# translate(config, CallTreeLocal, PostgresqlColumn, _.a + _.b)
translator = SqlTranslator.from_mappings(
        scalar, window, aggregate,
        PostgresqlColumn, PostgresqlColumnAgg
        )



"""
Implements LazyTbl to represent tables of SQL data, and registers it on verbs.

This module is responsible for the handling of the "table" side of things, while
translate.py handles translating column operations.


"""


from siuba.dply.verbs import (
        singledispatch2,
        show_query, collect,
        simple_varname,
        select, VarList, var_select,
        mutate,
        transmute,
        filter,
        arrange, _call_strip_ascending,
        summarize,
        count,
        group_by, ungroup,
        case_when,
        join, left_join, right_join, inner_join, semi_join, anti_join,
        head,
        rename,
        distinct,
        if_else
        )

from .translate import CustomOverClause, SqlColumn, SqlColumnAgg
from .utils import get_dialect_translator, _FixedSqlDatabase, _sql_select, _sql_column_collection, _sql_add_columns, _sql_with_only_columns

from sqlalchemy import sql
import sqlalchemy
from siuba.siu import Call, str_to_getitem_call, Lazy, FunctionLookupError
# TODO: currently needed for select, but can we remove pandas?
from pandas import Series

from sqlalchemy.sql import schema


# TODO:
#   - distinct
#   - annotate functions using sel.prefix_with("\n/*<Mutate Statement>*/\n") ?


# Helpers ---------------------------------------------------------------------

class SqlFunctionLookupError(FunctionLookupError): pass


class CallListener:
    """Generic listener. Each exit is called on a node's copy."""
    def enter(self, node):
        args, kwargs = node.map_subcalls(self.enter)

        return self.exit(node.__class__(node.func, *args, **kwargs))

    def exit(self, node):
        return node


class WindowReplacer(CallListener):
    """Call tree listener.

    Produces 2 important behaviors via the enter method:
      - returns evaluated sql call expression, with labels on all window expressions.
      - stores all labeled window expressions via the windows property.

    TODO: could replace with a sqlalchemy transformer
    """

    def __init__(self, columns, group_by, order_by, window_cte = None):
        self.columns = columns
        self.group_by = group_by
        self.order_by = order_by
        self.window_cte = window_cte
        self.windows = []

    def exit(self, node):
        col_expr = node(self.columns)

        if not isinstance(col_expr, sql.elements.ClauseElement):
            return col_expr

        over_clauses = [x for x in self._get_over_clauses(col_expr) if isinstance(x, CustomOverClause)]

        # put groupings and orderings onto custom over clauses
        for over in over_clauses:
            # TODO: shouldn't mutate these over clauses
            group_by = sql.elements.ClauseList(
                    *[self.columns[name] for name in self.group_by]
                    )
            order_by = sql.elements.ClauseList(
                    *_create_order_by_clause(self.columns, *self.order_by)
                    )

            over.set_over(group_by, order_by)

        if len(over_clauses) and self.window_cte is not None:
            # custom name, or parameters like "%(...)s" may nest and break psycopg2
            # with columns you can set a key to fix this, but it doesn't seem to 
            # be an option with labels
            name = self._get_unique_name('win', lift_inner_cols(self.window_cte))
            label = col_expr.label(name)

            # put into CTE, and return its resulting column, so that subsequent
            # operations will refer to the window column on window_cte. Note that
            # the operations will use the actual column, so may need to use the
            # ClauseAdaptor to make it a reference to the label
            self.window_cte = _sql_add_columns(self.window_cte, [label])
            win_col = lift_inner_cols(self.window_cte).values()[-1]
            self.windows.append(win_col)

            return win_col
                
        return col_expr
    
    @staticmethod
    def _get_unique_name(prefix, columns):
        column_names = set(columns.keys())

        i = 1
        name = prefix + str(i)
        while name in column_names:
            i += 1
            name = prefix + str(i)


        return name

    @staticmethod
    def _get_over_clauses(clause):
        windows = []
        append_win = lambda col: windows.append(col)

        sql.util.visitors.traverse(clause, {}, {"over": append_win})

        return windows


def track_call_windows(call, columns, group_by, order_by, window_cte = None):
    listener = WindowReplacer(columns, group_by, order_by, window_cte)
    col = listener.enter(call)
    return col, listener.windows, listener.window_cte


def lift_inner_cols(tbl):
    cols = list(tbl.inner_columns)
    data = {col.key: col for col in cols}

    return _sql_column_collection(data, cols)

def col_expr_requires_cte(call, sel, is_mutate = False):
    """Return whether a variable assignment needs a CTE"""

    call_vars = set(call.op_vars(attr_calls = False))

    sel_labs = get_inner_labels(sel)

    # I use the acronym fwg sol (frog soul) to remember sql clause eval order
    # from, where, group by, select, order by, limit
    # group clause evaluated before select clause, so not issue for mutate
    group_needs_cte = not is_mutate and len(sel._group_by_clause)
    
    return (   group_needs_cte
            # TODO: detect when a new var in mutate conflicts w/ order by
            #or len(sel._order_by_clause)
            or not sel_labs.isdisjoint(call_vars)
            )

def get_inner_labels(sel):
    columns = lift_inner_cols(sel)
    sel_labs = set(k for k,v in columns.items() if isinstance(v, sql.elements.Label))
    return sel_labs

def get_missing_columns(call, columns):
    missing_cols = set(call.op_vars(attr_calls = False)) - set(columns.keys())
    return missing_cols

def compile_el(tbl, el):
    compiled = el.compile(
         dialect = tbl.source.dialect,
         compile_kwargs = {"literal_binds": True}
    )
    return compiled

# Misc utilities --------------------------------------------------------------

def ordered_union(x, y):
    dx = {el: True for el in x}
    dy = {el: True for el in y}

    return tuple({**dx, **dy})




# Table -----------------------------------------------------------------------

class LazyTbl:
    def __init__(
            self, source, tbl, columns = None,
            ops = None, group_by = tuple(), order_by = tuple(),
            translator = None
            ):
        """Create a representation of a SQL table.

        Args:
            source: a sqlalchemy.Engine or sqlalchemy.Connection instance.
            tbl: table of form 'schema_name.table_name', 'table_name', or sqlalchemy.Table.
            columns: if specified, a listlike of column names.

        Examples
        --------

        ::
            from sqlalchemy import create_engine
            from siuba.data import mtcars

            # create database and table
            engine = create_engine("sqlite:///:memory:")
            mtcars.to_sql('mtcars', engine)

            tbl_mtcars = LazyTbl(engine, 'mtcars')
            
        """
        
        # connection and dialect specific functions
        self.source = sqlalchemy.create_engine(source) if isinstance(source, str) else source

        dialect = self.source.dialect.name
        self.translator = get_dialect_translator(dialect)

        self.tbl = self._create_table(tbl, columns, self.source)

        # important states the query can be in (e.g. grouped)
        self.ops = [self.tbl.select()] if ops is None else ops

        self.group_by = group_by
        self.order_by = order_by


    def append_op(self, op, **kwargs):
        cpy = self.copy(**kwargs)
        cpy.ops = cpy.ops + [op]
        return cpy

    def copy(self, **kwargs):
        return self.__class__(**{**self.__dict__, **kwargs})

    def shape_call(
            self,
            call, window = True, str_accessors = False,
            verb_name = None, arg_name = None,
            ):
        if isinstance(call, Call):
            pass
        elif str_accessors and isinstance(call, str):
            # verbs that can use strings as accessors, like group_by, or
            # arrange, need to convert those strings into a getitem call
            return str_to_getitem_call(call)
        elif isinstance(call, sql.elements.ColumnClause):
            return Lazy(call)
        elif callable(call):
            #TODO: should not happen here
            from siuba.siu import MetaArg
            return Call("__call__", call, MetaArg('_'))

        else:
            # verbs that use literal strings, need to convert them to a call
            # that returns a sqlalchemy "literal" object
            return Lazy(sql.literal(call))

        # raise informative error message if missing translation
        try:
            return self.translator.translate(call, window = window)
        except FunctionLookupError as err:
            raise SqlFunctionLookupError.from_verb(
                    verb_name or "Unknown",
                    arg_name or "Unknown",
                    err,
                    short = True
                    )


    def track_call_windows(self, call, columns = None, window_cte = None):
        """Returns tuple of (new column expression, list of window exprs)"""

        columns = self.last_op.columns if columns is None else columns
        return track_call_windows(call, columns, self.group_by, self.order_by, window_cte)

    def get_ordered_col_names(self):
        """Return columns from current select, with grouping columns first."""
        ungrouped = [k for k in self.last_op.columns.keys() if k not in self.group_by]
        return list(self.group_by) + ungrouped

    #def label_breaks_order_by(self, name):
    #    """Returns True if a new column label would break the order by vars."""

    #    # TODO: arrange currently allows literals, which breaks this. it seems
    #    #       better to only allow calls in arrange.
    #    order_by_vars = {c.op_vars(attr_calls=False) for c in self.order_by}




    @property
    def last_op(self):
        return self.ops[-1] if len(self.ops) else None

    @staticmethod
    def _create_table(tbl, columns = None, source = None):
        """Return a sqlalchemy.Table, autoloading column info if needed. 

        Arguments:
            tbl: a sqlalchemy.Table or string of form 'table_name' or 'schema_name.table_name'.
            columns: a tuple of column names for the table. Overrides source argument.
            source: a sqlalchemy engine, used to autoload columns.

        """
        if isinstance(tbl, sqlalchemy.Table):
            return tbl
        
        if not isinstance(tbl, str):
            raise ValueError("tbl must be a sqlalchemy Table or string, but was %s" %type(tbl))

        if columns is None and source is None:
            raise ValueError("One of columns or source must be specified")

        schema, table_name = tbl.split('.') if '.' in tbl else [None, tbl]

        columns = map(sqlalchemy.Column, columns) if columns is not None else tuple()

        # TODO: pybigquery uses schema to mean project_id, so we cannot use
        # siuba's classic breakdown "{schema}.{table_name}". Basically
        # pybigquery uses "{schema=project_id}.{dataset_dot_table_name}" in its internal
        # logic. An important side effect is that bigquery errors for
        # `dataset`.`table`, but not `dataset.table`.
        if source and source.dialect.name == "bigquery":
            table_name = tbl
            schema = None

        return sqlalchemy.Table(
                table_name,
                sqlalchemy.MetaData(bind = source),
                *columns,
                schema = schema,
                autoload_with = source if not columns else None
                )

    def _get_preview(self):
        # need to make prev op a cte, so we don't override any previous limit
        new_sel = self.last_op.alias().select().limit(5)
        tbl_small = self.append_op(new_sel)
        return collect(tbl_small)

    def __repr__(self):
        template = (
                "# Source: lazy query\n"
                "# DB Conn: {}\n"
                "# Preview:\n{}\n"
                "# .. may have more rows"
                )

        return template.format(repr(self.source.engine), repr(self._get_preview()))

    def _repr_html_(self):
        template = (
                "<div>"
                "<pre>"
                "# Source: lazy query\n"
                "# DB Conn: {}\n"
                "# Preview:\n"
                "</pre>"
                "{}"
                "<p># .. may have more rows</p>"
                "</div>"
                )

        data = self._get_preview()
        html_data = getattr(data, '_repr_html_', lambda: repr(data))()
        return template.format(self.source.engine, html_data)


def _repr_grouped_df_html_(self):
    return "<div><p>(grouped data frame)</p>" + self._selected_obj._repr_html_() + "</div>"



# Main Funcs 
# =============================================================================

# sql raw --------------

sql_raw = sql.literal_column

# show query -----------

from sqlalchemy.ext.compiler import compiles, deregister
from contextlib import contextmanager

@contextmanager
def use_simple_names():
    get_col_name = lambda el, *args, **kwargs: str(el.element.name)
    try:
        yield compiles(sql.compiler._CompileLabel)(get_col_name)
    except:
        pass
    finally:
        deregister(sql.compiler._CompileLabel)

@show_query.register(LazyTbl)
def _show_query(tbl, simplify = False):
    query = tbl.last_op #if not simplify else 
    compile_query = lambda: query.compile(
                dialect = tbl.source.dialect,
                compile_kwargs = {"literal_binds": True}
            )


    if simplify:
        # try to strip table names and labels where uneccessary
        with use_simple_names():
            print(compile_query())
    else:
        # use a much more verbose query
        print(compile_query())

    return tbl

# collect ----------

@collect.register(LazyTbl)
def _collect(__data, as_df = True):
    # TODO: maybe remove as_df options, always return dataframe
    # normally can just pass the sql objects to execute, but for some reason
    # psycopg2 completes about incomplete template.
    # see https://stackoverflow.com/a/47193568/1144523

    #query = __data.last_op
    #compiled = query.compile(
    #    dialect = __data.source.dialect,
    #    compile_kwargs = {"literal_binds": True}
    #)

    with __data.source.connect() as conn:
        if as_df:
            sql_db = _FixedSqlDatabase(conn)

            return sql_db.read_sql(__data.last_op)

        return conn.execute(__data.last_op)


@select.register(LazyTbl)
def _select(__data, *args, **kwargs):
    # see https://stackoverflow.com/questions/25914329/rearrange-columns-in-sqlalchemy-select-object
    if kwargs:
        raise NotImplementedError(
                "Using kwargs in select not currently supported. "
                "Use _.newname == _.oldname instead"
                )
    last_op = __data.last_op
    columns = {c.key: c for c in last_op.inner_columns}

    # same as for DataFrame
    colnames = Series(list(columns))
    vl = VarList()
    evaluated = (arg(vl) if callable(arg) else arg for arg in args)
    od = var_select(colnames, *evaluated)

    col_list = []
    for k,v in od.items():
        col = columns[k]
        col_list.append(col if v is None else col.label(v))

    return __data.append_op(last_op.with_only_columns(col_list))



@filter.register(LazyTbl)
def _filter(__data, *args):
    # TODO: aggregate funcs
    # Note: currently always produces 2 additional select statements,
    #       1 for window/aggs, and 1 for the where clause
    sel = __data.last_op.alias()                   # original select
    win_sel = sel.select()

    conds = []
    windows = []
    for ii, arg in enumerate(args):

        if isinstance(arg, Call):
            new_call = __data.shape_call(arg, verb_name = "Filter", arg_name = ii)
            #var_cols = new_call.op_vars(attr_calls = False)

            # note that a new win_sel is returned, w/ window columns appended
            col_expr, win_cols, win_sel = __data.track_call_windows(
                    new_call,
                    sel.columns,
                    window_cte = win_sel
                    )

            conds.append(col_expr)
            windows.extend(win_cols)
            
        else:
            conds.append(arg)

    bool_clause = sql.and_(*conds)

    # first cte, windows ----
    if len(windows):
        
        win_alias = win_sel.alias()

        # because track_call_windows in the loop above used select.append_column
        # multiple times, sqlalchemy doesn't know our window columns are the ones
        # in the final mutated for of win_sel
        #col_key_map = {col.key: col for col in win_alias.columns.values()}
        #equivalents = {col: [col_key_map[col.key]] for col in windows}

        # move non-window functions to refer to win_sel clause (not the innermost) ---
        bool_clause = sql.util.ClauseAdapter(win_alias) \
                .traverse(bool_clause)

        orig_cols = [win_alias.columns[k] for k in sel.columns.keys()]
    else:
        orig_cols = [sel]
    
    # create second cte ----
    filt_sel = _sql_select(orig_cols).where(bool_clause)
    return __data.append_op(filt_sel)


@mutate.register(LazyTbl)
def _mutate(__data, **kwargs):
    # Cases
    #  - work with group by
    #  - window functions
    # TODO: verify it can follow a renaming select

    # track labeled columns in set
    sel = __data.last_op

    # evaluate each call
    for colname, func in kwargs.items():
        # keep set of columns labeled (aliased) in this select statement
        # need to use inner cols, since sel.columns uses ColumnClause, not Label
        labs = set(k for k,v in lift_inner_cols(sel).items() if isinstance(v, sql.elements.Label))
        new_call = __data.shape_call(func, verb_name = "Mutate", arg_name = colname)

        sel = _mutate_select(sel, colname, new_call, labs, __data)

    return __data.append_op(sel)


def _mutate_select(sel, colname, func, labs, __data):
    """Return select statement containing a new column, func expr as colname.
    
    Note: since a func can refer to previous columns generated by mutate, this
    function handles whether to add a column to the existing select statement,
    or to use it as a subquery.
    """
    replace_col = colname in lift_inner_cols(sel)
    # Call objects let us check whether column expr used a derived column
    # e.g. SELECT a as b, b + 1 as c raises an error in SQL, so need subquery
    if not col_expr_requires_cte(func, sel, is_mutate = True):
        # New column may be able to modify existing select
        columns = lift_inner_cols(sel)

    else:
        # anything else requires a subquery
        cte = sel.alias(None)
        columns = cte.columns
        sel = cte.select()

    # evaluate call expr on columns, making sure to use group vars
    new_col, windows, _ = __data.track_call_windows(func, columns)

    # replacing an existing column, so strip it from select statement
    if replace_col:
        replaced = {**columns}
        replaced[colname] = new_col.label(colname)
        return sel.with_only_columns(list(replaced.values()))

    return _sql_add_columns(sel, [new_col.label(colname)])


@transmute.register(LazyTbl)
def _transmute(__data, **kwargs):
    # will use mutate, then select some cols
    f_mutate = mutate.registry[type(__data)]

    # transmute keeps grouping cols, and any defined in kwargs
    cols_to_keep = ordered_union(__data.group_by, kwargs)

    sel = f_mutate(__data, **kwargs).last_op

    columns = lift_inner_cols(sel)
    sel_stripped = sel.with_only_columns([columns[k] for k in cols_to_keep])

    return __data.append_op(sel_stripped)


@arrange.register(LazyTbl)
def _arrange(__data, *args):
    # Note that SQL databases often do not subquery order by clauses. Arrange
    # sets order_by on the backend, so it can set order by in over elements,
    # and handle when new columns are named the same as order by vars.
    # see: https://dba.stackexchange.com/q/82930

    last_op = __data.last_op
    cols = lift_inner_cols(last_op)

    
    new_calls = []
    for ii, expr in enumerate(args):
        if callable(expr):

            res = __data.shape_call(
                    expr, window = False,
                    verb_name = "Arrange", arg_name = ii
                    )

        else:
            res = expr

        new_calls.append(res)

    sort_cols = _create_order_by_clause(cols, *new_calls)

    order_by = __data.order_by + tuple(new_calls)
    return __data.append_op(last_op.order_by(*sort_cols), order_by = order_by)


# TODO: consolidate / pull expr handling funcs into own file?
def _create_order_by_clause(columns, *args):
    sort_cols = []
    for arg in args:
        # simple named column
        if isinstance(arg, str):
            sort_cols.append(columns[arg])
        # an expression
        elif callable(arg):
            # handle special case where -_.colname -> colname DESC
            f, asc = _call_strip_ascending(arg)
            col_op = f(columns) if asc else f(columns).desc()
            #col_op = arg(columns)
            sort_cols.append(col_op)
        else:
            raise NotImplementedError("Must be string or callable")

    return sort_cols



@count.register(LazyTbl)
def _count(__data, *args, sort = False, wt = None, **kwargs):
    # TODO: if already col named n, use name nn, etc.. get logic from tidy.py
    if wt is not None:
        raise NotImplementedError("TODO")

    res_name = "n"
    # similar to filter verb, we need two select statements,
    # an inner one for derived cols, and outer to group by them

    # inner select ----
    # holds any mutation style columns
    arg_names = []
    for arg in args:
        name = simple_varname(arg)
        if name is None:
            raise NotImplementedError(
                    "Count positional arguments must be single column name. "
                    "Use a named argument to count using complex expressions."
                    )
        arg_names.append(name)

    tbl_inner = mutate(__data, **kwargs)
    sel_inner = tbl_inner.last_op
    group_cols = arg_names + list(kwargs)

    # create outer select ----
    # holds selected columns and tally (n)
    sel_inner_cte = sel_inner.alias()
    inner_cols = sel_inner_cte.columns

    # apply any group vars from a group_by verb call first
    tbl_group_cols = [inner_cols[k] for k in tbl_inner.group_by]
    count_group_cols = [inner_cols[k] for k in group_cols]

    # combine with any defined in the count verb call
    outer_group_cols = ordered_union(tbl_group_cols, count_group_cols)
    # holds the actual count (e.g. n)
    count_col = sql.functions.count().label(res_name)

    sel_outer = _sql_select([*outer_group_cols, count_col]) \
            .select_from(sel_inner_cte) \
            .group_by(*outer_group_cols)

    # count is like summarize, so removes order_by
    return tbl_inner.append_op(
            sel_outer.order_by(count_col.desc()),
            order_by = tuple()
            )


@summarize.register(LazyTbl)
def _summarize(__data, **kwargs):
    # https://stackoverflow.com/questions/14754994/why-is-sqlalchemy-count-much-slower-than-the-raw-query
    # what if windowed mutate or filter has been done? 
    #   - filter is fine, since it uses a CTE
    #   - need to detect any window functions...
    old_sel = __data.last_op._clone()

    new_calls = {}
    for k, expr in kwargs.items():
        new_calls[k] = __data.shape_call(
                expr, window = False,
                verb_name = "Summarize", arg_name = k
                )

    needs_cte = [col_expr_requires_cte(call, old_sel) for call in new_calls.values()]
    group_on_labels = set(__data.group_by) & get_inner_labels(old_sel)

    # create select statement ----

    if any(needs_cte) or len(group_on_labels):
        # need a cte, due to alias cols or existing group by
        # current select stmt has group by clause, so need to make it subquery
        cte = old_sel.alias()
        columns = cte.columns
        sel = sql.select().select_from(cte)
    else:
        # otherwise, can alter the existing select statement
        columns = lift_inner_cols(old_sel)
        sel = old_sel

        # explicitly add original from clause tables, since we will be limiting
        # the columns this select uses, which causes sqlalchemy to remove
        # unreferenced tables
        for _from in sel.froms:
            sel = sel.select_from(_from)
        

    # add group by columns ----
    group_cols = [columns[k] for k in __data.group_by]

    # add each aggregate column ----
    # TODO: can't do summarize(b = mean(a), c = b + mean(a))
    #       since difficult for c to refer to agg and unagg cols in SQL
    expr_cols = []
    for k, expr in new_calls.items():
        missing_cols = get_missing_columns(expr, columns)
        if missing_cols:
            raise NotImplementedError(
                    "Summarize cannot find the following columns: %s. "
                    "Note that it cannot refer to variables defined earlier in the "
                    "same summarize call." % missing_cols
                    )

        col = expr(columns).label(k)
        expr_cols.append(col)
        

    all_cols = [*group_cols, *expr_cols]
    final_sel = _sql_with_only_columns(sel, all_cols).group_by(*group_cols)
    new_data = __data.append_op(final_sel, group_by = tuple(), order_by = tuple())
    return new_data


@group_by.register(LazyTbl)
def _group_by(__data, *args, add = False, **kwargs):
    if kwargs:
        data = mutate(__data, **kwargs)
    else:
        data = __data

    # put kwarg grouping vars last, so similar order to function call
    groups =  tuple(simple_varname(arg) for arg in args) + tuple(kwargs)
    if None in groups:
        raise NotImplementedError("Complex expressions not supported in sql group_by")

    # ensure group_by variables are in the select columns
    cols = data.last_op.alias().columns
    unmatched = set(groups) - set(cols.keys())
    if unmatched:
        raise KeyError("group_by specifies columns missing from table: %s" %unmatched)

    if add:
        groups = ordered_union(data.group_by, groups)

    return data.copy(group_by = groups)


@ungroup.register(LazyTbl)
def _ungroup(__data):
    return __data.copy(group_by = tuple())


@case_when.register(sql.base.ImmutableColumnCollection)
def _case_when(__data, cases):
    # TODO: will need listener to enter case statements, to handle when they use windows
    if isinstance(cases, Call):
        cases = cases(__data)

    whens = []
    case_items = list(cases.items())
    n_items = len(case_items)

    else_val = None
    for ii, (expr, val) in enumerate(case_items):
        # handle where val is a column expr
        if callable(val):
            val = val(__data)

        # handle when expressions
        if ii+1 == n_items and expr is True:
            else_val = val
        elif callable(expr):
            whens.append((expr(__data), val))
        else:
            whens.append((expr, val))

    return sql.case(whens, else_ = else_val)
        

# Join ------------------------------------------------------------------------

from collections.abc import Mapping

def _joined_cols(left_cols, right_cols, on_keys, how, suffix = ("_x", "_y")):
    """Return labeled columns, according to selection rules for joins.

    Rules:
        1. For join keys, keep left table's column
        2. When keys have the same labels, add suffix
    """

    # TODO: remove sets, so uses stable ordering
    # when left and right cols have same name, suffix with _x / _y
    keep_right = set(right_cols.keys()) - set(on_keys.values())
    shared_labs = set(left_cols.keys()).intersection(keep_right)

    right_cols_no_keys = {k: right_cols[k] for k in keep_right}

    # for an outer join, have key columns coalesce values

    left_cols = {**left_cols}
    if how == "full":
        for lk, rk in on_keys.items():
            col = sql.functions.coalesce(left_cols[lk], right_cols[rk])
            left_cols[lk] = col.label(lk)
    elif how == "right":
        for lk, rk in on_keys.items():
            # Make left key columns actually be right ones (which contain left + extra)
            left_cols[lk] = right_cols[rk].label(lk)


    # create labels ----
    l_labs = _relabeled_cols(left_cols, shared_labs, suffix[0])
    r_labs = _relabeled_cols(right_cols_no_keys, shared_labs, suffix[1])

    return l_labs + r_labs
    


def _relabeled_cols(columns, keys, suffix):
    # add a suffix to all columns with names in keys
    cols = []
    for k, v in columns.items():
        new_col = v.label(k + str(suffix)) if k in keys else v
        cols.append(new_col)
    return cols


@join.register(LazyTbl)
def _join(left, right, on = None, *args, how = "inner", sql_on = None):
    _raise_if_args(args)

    # Needs to be on the table, not the select
    left_sel = left.last_op.alias()
    right_sel = right.last_op.alias()

    # handle arguments ----
    on  = _validate_join_arg_on(on, sql_on)
    how = _validate_join_arg_how(how)
    
    # for equality join used to combine keys into single column
    consolidate_keys = on if sql_on is None else {}
    
    if how == "right":
        # switch joins, since sqlalchemy doesn't have right join arg
        # see https://stackoverflow.com/q/11400307/1144523
        left_sel, right_sel = right_sel, left_sel
        on = {v:k for k,v in on.items()}

    # create join conditions ----
    bool_clause = _create_join_conds(left_sel, right_sel, on)

    # create join ----
    join = left_sel.join(
            right_sel,
            onclause = bool_clause,
            isouter = how != "inner",
            full = how == "full"
            )

    # if right join, set selects back
    if how == "right":
        left_sel, right_sel = right_sel, left_sel
        on = {v:k for k,v in on.items()}

    # note, shared_keys assumes on is a mapping...
    # TODO: shared_keys appears to be for when on is not specified, but was unused
    #shared_keys = [k for k,v in on.items() if k == v]
    labeled_cols = _joined_cols(
            left_sel.columns,
            right_sel.columns,
            on_keys = consolidate_keys,
            how = how
            )

    sel = _sql_select(labeled_cols).select_from(join)
    return left.append_op(sel, order_by = tuple())


@semi_join.register(LazyTbl)
def _semi_join(left, right = None, on = None, *args, sql_on = None):
    _raise_if_args(args)

    left_sel = left.last_op.alias()
    right_sel = right.last_op.alias()

    # handle arguments ----
    on  = _validate_join_arg_on(on, sql_on)
    
    # create join conditions ----
    bool_clause = _create_join_conds(left_sel, right_sel, on)

    # create inner join ----
    exists_clause = _sql_select([sql.literal(1)]) \
            .select_from(right_sel) \
            .where(bool_clause)

    # only keep left hand select's columns ----
    sel = _sql_select(left_sel.columns) \
            .select_from(left_sel) \
            .where(sql.exists(exists_clause))

    return left.append_op(sel, order_by = tuple())


@anti_join.register(LazyTbl)
def _anti_join(left, right = None, on = None, *args, sql_on = None):
    _raise_if_args(args)

    left_sel = left.last_op.alias()
    right_sel = right.last_op.alias()

    # handle arguments ----
    on  = _validate_join_arg_on(on, sql_on)
    
    # create join conditions ----
    bool_clause = _create_join_conds(left_sel, right_sel, on)

    # create inner join ----
    #not_exists = ~sql.exists([1], from_obj = right_sel).where(bool_clause)
    exists_clause = _sql_select([sql.literal(1)]) \
            .select_from(right_sel) \
            .where(bool_clause)

    sel = left_sel.select().where(~sql.exists(exists_clause))
    
    return left.append_op(sel, order_by = tuple())
       
def _raise_if_args(args):
    if len(args):
        raise NotImplemented("*args is reserved for future arguments (e.g. suffix)")

def _validate_join_arg_on(on, sql_on = None):
    # handle sql on case
    if sql_on is not None:
        if on is not None:
            raise ValueError("Cannot specify both on and sql_on")

        return sql_on

    # handle general cases
    if on is None:
        raise NotImplementedError("on arg currently cannot be None (default) for SQL")
    elif isinstance(on, str):
        on = {on: on}
    elif isinstance(on, (list, tuple)):
        on = dict(zip(on, on))

    if not isinstance(on, Mapping):
        raise TypeError("on must be a Mapping (e.g. dict)")

    return on

def _validate_join_arg_how(how):
    how_options = ("inner", "left", "right", "full")
    if how not in how_options:
        raise ValueError("how argument needs to be one of %s" %how_options)
    
    return how

def _create_join_conds(left_sel, right_sel, on):
    left_cols  = left_sel.columns  #lift_inner_cols(left_sel)
    right_cols = right_sel.columns #lift_inner_cols(right_sel)

    if callable(on):
        # callable, like with sql_on arg
        conds = [on(left_cols, right_cols)]
    else:
        # dict-like of form {left: right}
        conds = []
        for l, r in on.items():
            col_expr = left_cols[l] == right_cols[r]
            conds.append(col_expr)
            
    return sql.and_(*conds)
    

# Head ------------------------------------------------------------------------

@head.register(LazyTbl)
def _head(__data, n = 5):
    sel = __data.last_op
    
    return __data.append_op(sel.limit(n))


# Rename ----------------------------------------------------------------------

@rename.register(LazyTbl)
def _rename(__data, **kwargs):
    sel = __data.last_op
    columns = lift_inner_cols(sel)

    # old_keys uses dict as ordered set
    old_to_new = {simple_varname(v):k for k,v in kwargs.items()}
    
    if None in old_to_new:
        raise KeyError("positional arguments must be simple column, "
                        "e.g. _.colname or _['colname']"
                        )

    labs = [c.label(old_to_new[k]) if k in old_to_new else c for k,c in columns.items()]

    new_sel = sel.with_only_columns(labs)

    return __data.append_op(new_sel)


# Distinct --------------------------------------------------------------------

@distinct.register(LazyTbl)
def _distinct(__data, *args, _keep_all = False, **kwargs):
    if (args or kwargs) and _keep_all:
        raise NotImplementedError("Distinct with variables specified in sql requires _keep_all = False")
    
    inner_sel = mutate(__data, **kwargs).last_op if kwargs else __data.last_op

    # TODO: this is copied from the df distinct version
    # cols dict below is used as ordered set
    cols = {simple_varname(x): True for x in args}
    cols.update(kwargs)

    if None in cols:
        raise KeyError("positional arguments must be simple column, "
                        "e.g. _.colname or _['colname']"
                        )

    # use all columns by default
    if not cols:
        cols = lift_inner_cols(inner_sel).keys()

    if not len(inner_sel._order_by_clause):
        # select distinct has to include any columns in the order by clause,
        # so can only safely modify existing statement when there's no order by
        sel_cols = lift_inner_cols(inner_sel)
        distinct_cols = [sel_cols[k] for k in cols]
        sel = inner_sel.with_only_columns(distinct_cols).distinct()
    else:
        # fallback to cte
        cte = inner_sel.alias()
        distinct_cols = [cte.columns[k] for k in cols]
        sel = _sql_select(distinct_cols).select_from(cte).distinct()

    return __data.append_op(sel)

    
# if_else ---------------------------------------------------------------------

@if_else.register(sql.elements.ColumnElement)
def _if_else(cond, true_vals, false_vals):
    whens = [(cond, true_vals)]
    return sql.case(whens, else_ = false_vals)





from .verbs import LazyTbl, sql_raw, SqlFunctionLookupError
from .translate import SqlColumn, SqlColumnAgg

# preceed w/ underscore so it isn't exported by default
# we just want to register the singledispatch funcs
from .dply import vector as _vector
from .dply import string as _string



from ..translate import SqlTranslator

from .postgresql import scalar, aggregate, window, funcs, PostgresqlColumn, PostgresqlColumnAgg

class RedshiftColumn(PostgresqlColumn): pass
class RedshiftColumnAgg(PostgresqlColumnAgg): pass

translator = SqlTranslator.from_mappings(
        scalar, window, aggregate,
        RedshiftColumn, RedshiftColumnAgg
        )
